"""Base class for acquisition optimizers."""

import abc
import dataclasses
from typing import Any, Callable, List, Optional

import attr
from vizier import pyvizier as vz
from vizier.pyvizier import converters

Array = Any


def _is_positive(instance, attribute, value):
  del instance, attribute
  if value < 0:
    raise ValueError(f'value must be positive! given: {value}')


@attr.s(frozen=False, init=True, slots=True)
class BranchSelection:
  """Branch Selection.

  Decision of a branch selector that contains the flat subspace and number
  of suggestions to be generated in that subspace.

  Instead of creating N suggestions on a conditional space S, we create
  N_1, ..., N_k suggestions on flat spaces S_1, .., S_k such that
  N_1 + ... + N_k = N and S_1, ..., S_k are contained in S. Each BranchSelection
  object represents (N_i, S_i) tuple.

  Attributres:
    search_space: Search space that does not contain any conditional parameters.
    num_suggestions: Number of suggestions to be made in the search space.
  """
  search_space: vz.SearchSpace = attr.ib(
      validator=attr.validators.instance_of(vz.SearchSpace),
      on_setattr=attr.setters.validate)
  num_suggestions: int = attr.ib(
      validator=[attr.validators.instance_of(int), _is_positive],
      on_setattr=attr.setters.validate)


class BranchSelector(abc.ABC):

  @abc.abstractmethod
  def select_branches(self, num_suggestions: int) -> List[BranchSelection]:
    pass


class GradientFreeMaximizer(abc.ABC):
  """Optimizes a function on Vizier search space.

  Typically used for optimizing acquisition functions.
  """

  @abc.abstractmethod
  def maximize(self,
               score_fn: Callable[[Array], Array],
               converter: converters.DefaultTrialConverter,
               search_space: vz.SearchSpace,
               *,
               count: int = 1,
               budget_factor: float = 1.0,
               **kwargs) -> List[vz.Trial]:
    """Maximize a function.

    Args:
      score_fn: A function that takes as input `converter.to_features()` and
        outputs a numpy array-like object of shape (B,).
      converter: Responsible for converting between `Trial`s (which is the
        return type) and `score_fn`'s desired input type.
      search_space: Returned Trials must be contained in this search space.
      count: Optimizer tries to return this many trials.
      budget_factor: Every optimizer has a rough notion of "standard" budget.
        Use this much fraction of the standard budget for the call.
      **kwargs: For experimental keyword arguments.

    Returns:
      Trials, of length less than or equal to max_num_suggestions.
    """
    pass


@dataclasses.dataclass(frozen=True)
class BranchThenMaximizer(GradientFreeMaximizer):
  """Maximizes a function by first choosing a branch and then apply maximizer.

  Attributes:
    branch_selector: Selects all conditional parent values
    maximizer_factory: Creates an optimizer in the flat (non-conditional) after
      branch selector fixing all conditional parent values.
    max_num_suggestions_per_branch: Limits the number of suggestions per branch.
      This is useful when acquisition function isn't well-suited for batch
      suggestions.
  """
  branch_selector: BranchSelector
  maximizer_factory: Callable[[], GradientFreeMaximizer]
  max_num_suggestions_per_branch: Optional[int] = None

  def _num_suggestions_for_branch(self, branch: BranchSelection) -> int:
    if self.max_num_suggestions_per_branch is None:
      return branch.num_suggestions
    else:
      return min(self.max_num_suggestions_per_branch, branch.num_suggestions)

  def maximize(self,
               score_fn: Callable[[Array], Array],
               converter: converters.DefaultTrialConverter,
               search_space: vz.SearchSpace,
               *,
               count: int = 1,
               budget_factor: float = 1.0,
               **kwargs) -> List[vz.Trial]:
    # If there are conditional branches, use Vizier's default branch
    # selection mechanism.
    branches = self.branch_selector.select_branches(count)
    suggestions = []
    maximizer = self.maximizer_factory()
    for branch in branches:
      suggestions.extend(
          maximizer.maximize(
              score_fn,
              converter,
              branch.search_space,
              count=self._num_suggestions_for_branch(branch),
              budget_factor=budget_factor * (branch.num_suggestions / count)))
    return suggestions



"""Convenience classes for configuring Vizier Early-Stopping Configs."""
import copy
from typing import Union

import attr
from vizier.service import study_pb2

AutomatedStoppingConfigProto = Union[
    study_pb2.StudySpec.DecayCurveAutomatedStoppingSpec,
    study_pb2.StudySpec.MedianAutomatedStoppingSpec]


@attr.s(frozen=True, init=True, slots=True, kw_only=True)
class AutomatedStoppingConfig:
  """A wrapper for study_pb2.automated_stopping_spec."""
  _proto: AutomatedStoppingConfigProto = attr.ib(init=True, kw_only=True)

  @classmethod
  def decay_curve_stopping_config(cls,
                                  use_steps: bool) -> 'AutomatedStoppingConfig':
    """Create a DecayCurve automated stopping config.

    Vizier will early stop the Trial if it predicts the Trial objective value
    will not be better than previous Trials.

    Args:
      use_steps: Bool. If set, use Measurement.step_count as the measure of
        training progress.  Otherwise, use Measurement.elapsed_duration.

    Returns:
      AutomatedStoppingConfig object.

    Raises:
      ValueError: If more than one metric is configured.
      Note that Vizier Early Stopping currently only supports single-objective
      studies.
    """
    config = study_pb2.StudySpec.DecayCurveAutomatedStoppingSpec(
        use_elapsed_duration=not use_steps)
    return cls(proto=config)

  @classmethod
  def median_automated_stopping_config(
      cls, use_steps: bool) -> 'AutomatedStoppingConfig':
    """Create a Median automated stopping config.

    Vizier will early stop the Trial if it predicts the Trial objective value
    will not be better than previous Trials.

    Args:
      use_steps: Bool. If set, use Measurement.step_count as the measure of
        training progress.  Otherwise, use Measurement.elapsed_duration.

    Returns:
      AutomatedStoppingConfig object.

    Raises:
      ValueError: If more than one metric is configured.
      Note that Vizier Early Stopping currently only supports single-objective
      studies.
    """
    config = study_pb2.StudySpec.MedianAutomatedStoppingSpec(
        use_elapsed_duration=not use_steps)
    return cls(proto=config)

  @classmethod
  def from_proto(
      cls, proto: AutomatedStoppingConfigProto) -> 'AutomatedStoppingConfig':
    return cls(proto=proto)

  def to_proto(self) -> AutomatedStoppingConfigProto:
    """Returns this object as a proto."""
    return copy.deepcopy(self._proto)



"""Tests for vizier.service.service_policy_supporter."""

import logging
from vizier import pythia
from vizier.service import pyvizier
from vizier.service import resources
from vizier.service import service_policy_supporter
from vizier.service import study_pb2
from vizier.service import test_util
from vizier.service import vizier_server

from absl.testing import absltest


class PythiaSupporterTest(absltest.TestCase):

  def setUp(self):
    self.owner_id = 'my_username'
    self.study_id = '1231223'
    self.study_name = resources.StudyResource(
        owner_id=self.owner_id, study_id=self.study_id).name
    self.vs = vizier_server.VizierService()
    self.example_study = test_util.generate_study(self.owner_id, self.study_id)
    self.vs.datastore.create_study(self.example_study)

    self.active_trials = test_util.generate_trials(
        range(1, 7), self.owner_id, self.study_id)
    self.succeeded_trial = test_util.generate_trials(
        [7],
        self.owner_id,
        self.study_id,
        state=study_pb2.Trial.State.SUCCEEDED,
        final_measurement=study_pb2.Measurement())[0]

    for trial in self.active_trials + [self.succeeded_trial]:
      self.vs.datastore.create_trial(trial)

    self.policy_supporter = service_policy_supporter.ServicePolicySupporter(
        self.study_name, self.vs)

    super().setUp()

  def test_trial_names_filter(self):
    trials = self.policy_supporter.GetTrials(
        study_guid=self.study_name, trial_ids=[3, 4])

    self.assertEqual(trials[0],
                     pyvizier.TrialConverter.from_proto(self.active_trials[2]))
    self.assertEqual(trials[1],
                     pyvizier.TrialConverter.from_proto(self.active_trials[3]))

  def test_min_max_filter(self):
    trials = self.policy_supporter.GetTrials(
        study_guid=self.study_name, min_trial_id=3, max_trial_id=4)

    self.assertEqual(trials[0],
                     pyvizier.TrialConverter.from_proto(self.active_trials[2]))
    self.assertEqual(trials[1],
                     pyvizier.TrialConverter.from_proto(self.active_trials[3]))

  def test_status_match_filter(self):
    trials = self.policy_supporter.GetTrials(
        study_guid=self.study_name, status_matches=pyvizier.TrialStatus.ACTIVE)

    self.assertLen(trials, 6)
    self.assertEqual(trials[0],
                     pyvizier.TrialConverter.from_proto(self.active_trials[0]))

  def test_raise_value_error(self):

    def should_raise_value_error_fn():
      self.policy_supporter.GetTrials(
          study_guid=self.study_name,
          trial_ids=[1],
          status_matches=pyvizier.TrialStatus.ACTIVE)

      with self.assertRaises(ValueError):
        should_raise_value_error_fn()

  def test_get_study_config(self):
    pythia_sc = self.policy_supporter.GetStudyConfig(self.study_name)
    correct_pythia_sc = pyvizier.StudyConfig.from_proto(
        self.example_study.study_spec)
    self.assertEqual(pythia_sc, correct_pythia_sc)

  def test_update_metadata(self):
    on_study_metadata = pyvizier.Metadata()
    on_study_metadata.ns('bar')['foo'] = '.bar.foo.1'
    on_trial1_metadata = pyvizier.Metadata()
    on_trial1_metadata.ns('bax')['nerf'] = '1.bar.nerf.2'
    delta = pythia.MetadataDelta(
        on_study=on_study_metadata, on_trials={1: on_trial1_metadata})
    self.policy_supporter.SendMetadata(delta)
    # Read to see that the results are correct.
    pythia_sc = self.policy_supporter.GetStudyConfig(self.study_name)
    self.assertLen(pythia_sc.metadata.namespaces(), 1)
    logging.info('pythia_sc namespaces: %s', pythia_sc.metadata.namespaces())
    logging.info('pythia_sc = %s', pythia_sc.metadata.ns('bar'))
    logging.info('on_study = %s', on_study_metadata.ns('bar'))
    self.assertNotEmpty(pythia_sc.metadata.ns('bar'))
    self.assertEqual(pythia_sc.metadata.ns('bar'), on_study_metadata.ns('bar'))
    trials = self.policy_supporter.GetTrials(
        study_guid=self.study_name, min_trial_id=1, max_trial_id=1)
    self.assertLen(trials, 1)
    self.assertNotEmpty(trials[0].metadata.ns('bax'))
    self.assertEqual(trials[0].metadata.ns('bax'), on_trial1_metadata.ns('bax'))


if __name__ == '__main__':
  absltest.main()



"""Json utils."""

import json
from typing import Any

import numpy as np


class NumpyEncoder(json.JSONEncoder):
  """Example: json.dumps(np.array(...), cls=NumpyEncoder)."""

  def default(self, obj: Any) -> Any:
    if isinstance(obj, np.ndarray):
      # Shape must be dumped and restored, in case that the array has shape
      # that includes 0-sized dimensions.
      return {
          'dtype': np.dtype(obj.dtype).name,
          'value': obj.tolist(),
          'shape': obj.shape
      }
    return json.JSONEncoder.default(self, obj)


def numpy_hook(obj: Any) -> Any:
  """Example: json.loads(..., object_hook=numpy_hook)."""
  if 'dtype' not in obj:
    return obj
  if 'shape' not in obj:
    return obj
  return np.array(obj['value'], dtype=obj['dtype']).reshape(obj['shape'])


class NumpyDecoder(json.JSONDecoder):
  """Convert list to numpy if we see a list during json decoding.

  Example: json.loads(json_string, cls=NumpyDecoder).
  """

  def __init__(self, *args, **kargs):
    super(NumpyDecoder, self).__init__(object_hook=numpy_hook, *args, **kargs)



"""Tests for vizier.pyvizier.shared.base_study_config."""

import numpy as np
from vizier._src.pyvizier.shared import base_study_config
from vizier._src.pyvizier.shared import parameter_config as pc
from vizier._src.pyvizier.shared import trial
from absl.testing import absltest
from absl.testing import parameterized


class ObjectiveMetricGoalTest(absltest.TestCase):

  def test_basics(self):
    self.assertTrue(base_study_config.ObjectiveMetricGoal.MAXIMIZE.is_maximize)
    self.assertFalse(base_study_config.ObjectiveMetricGoal.MAXIMIZE.is_minimize)
    self.assertTrue(base_study_config.ObjectiveMetricGoal.MINIMIZE.is_minimize)
    self.assertFalse(base_study_config.ObjectiveMetricGoal.MINIMIZE.is_maximize)


class MetricTypeTest(absltest.TestCase):

  def test_basics(self):
    self.assertTrue(base_study_config.MetricType.SAFETY.is_safety)
    self.assertTrue(base_study_config.MetricType.OBJECTIVE.is_objective)


class MetricInformationTest(absltest.TestCase):

  def testMinMaxValueDefault(self):
    info = base_study_config.MetricInformation(
        goal=base_study_config.ObjectiveMetricGoal.MAXIMIZE)
    self.assertEqual(info.min_value, -np.inf)
    self.assertEqual(info.max_value, np.inf)

  def testMinMaxValueSet(self):
    info = base_study_config.MetricInformation(
        goal=base_study_config.ObjectiveMetricGoal.MAXIMIZE,
        min_value=-1.,
        max_value=1.)
    self.assertEqual(info.min_value, -1.)
    self.assertEqual(info.max_value, 1.)

  def testMinMaxBadValueInit(self):
    with self.assertRaises(ValueError):
      base_study_config.MetricInformation(
          goal=base_study_config.ObjectiveMetricGoal.MAXIMIZE,
          min_value=1.,
          max_value=-1.)

  def testMinMaxBadValueSet(self):
    info = base_study_config.MetricInformation(
        goal=base_study_config.ObjectiveMetricGoal.MAXIMIZE,
        min_value=-1.,
        max_value=1.)
    with self.assertRaises(ValueError):
      info.min_value = 2.
    with self.assertRaises(ValueError):
      info.max_value = -2.


class MetricsConfigTest(parameterized.TestCase):

  def testBasics(self):
    config = base_study_config.MetricsConfig()
    config.append(
        base_study_config.MetricInformation(
            name='max1', goal=base_study_config.ObjectiveMetricGoal.MAXIMIZE))
    config.extend([
        base_study_config.MetricInformation(
            name='max_safe1',
            goal=base_study_config.ObjectiveMetricGoal.MAXIMIZE,
            safety_threshold=0.0),
        base_study_config.MetricInformation(
            name='max2', goal=base_study_config.ObjectiveMetricGoal.MAXIMIZE),
        base_study_config.MetricInformation(
            name='min1', goal=base_study_config.ObjectiveMetricGoal.MINIMIZE),
        base_study_config.MetricInformation(
            name='min_safe2',
            goal=base_study_config.ObjectiveMetricGoal.MINIMIZE,
            safety_threshold=0.0)
    ])
    self.assertLen(config, 5)
    self.assertLen(config.of_type(base_study_config.MetricType.OBJECTIVE), 3)
    self.assertLen(config.of_type(base_study_config.MetricType.SAFETY), 2)

  def testDuplicateNames(self):
    config = base_study_config.MetricsConfig()
    config.append(
        base_study_config.MetricInformation(
            name='max1', goal=base_study_config.ObjectiveMetricGoal.MAXIMIZE))
    with self.assertRaises(ValueError):
      config.append(
          base_study_config.MetricInformation(
              name='max1', goal=base_study_config.ObjectiveMetricGoal.MAXIMIZE))


class SearchSpaceTest(parameterized.TestCase):

  def testAddFloatParamMinimal(self):
    space = base_study_config.SearchSpace()
    self.assertEmpty(space.parameters)
    selector = space.select_root().add_float_param('f1', 1.0, 15.0)
    # Test the returned selector.
    self.assertEqual(selector.path_string, '')
    self.assertEqual(selector.parameter_name, 'f1')
    self.assertEqual(selector.parameter_values, [])
    # Test the search space.
    self.assertLen(space.parameters, 1)
    self.assertEqual(space.parameters[0].name, 'f1')
    self.assertEqual(space.parameters[0].type, pc.ParameterType.DOUBLE)
    self.assertEqual(space.parameters[0].bounds, (1.0, 15.0))
    self.assertEqual(space.parameters[0].scale_type, pc.ScaleType.LINEAR)
    self.assertEmpty(space.parameters[0].matching_parent_values)
    self.assertEmpty(space.parameters[0].child_parameter_configs)
    with self.assertRaisesRegex(ValueError, 'feasible_values is invalid.*'):
      _ = space.parameters[0].feasible_values
    self.assertIsNone(space.parameters[0].default_value)

    _ = space.select_root().add_float_param('f2', 2.0, 16.0)
    self.assertLen(space.parameters, 2)
    self.assertEqual(space.parameters[0].name, 'f1')
    self.assertEqual(space.parameters[0].type, pc.ParameterType.DOUBLE)
    self.assertEqual(space.parameters[0].bounds, (1.0, 15.0))
    self.assertEqual(space.parameters[1].name, 'f2')
    self.assertEqual(space.parameters[1].type, pc.ParameterType.DOUBLE)
    self.assertEqual(space.parameters[1].bounds, (2.0, 16.0))

  def testAddFloatParam(self):
    space = base_study_config.SearchSpace()
    self.assertEmpty(space.parameters)
    _ = space.select_root().add_float_param(
        'f1', 1.0, 15.0, default_value=3.0, scale_type=pc.ScaleType.LOG)
    self.assertLen(space.parameters, 1)
    self.assertEqual(space.parameters[0].name, 'f1')
    self.assertEqual(space.parameters[0].type, pc.ParameterType.DOUBLE)
    self.assertEqual(space.parameters[0].bounds, (1.0, 15.0))
    self.assertEqual(space.parameters[0].scale_type, pc.ScaleType.LOG)
    self.assertEmpty(space.parameters[0].matching_parent_values)
    self.assertEmpty(space.parameters[0].child_parameter_configs)
    with self.assertRaisesRegex(ValueError, 'feasible_values is invalid.*'):
      _ = space.parameters[0].feasible_values
    self.assertEqual(space.parameters[0].default_value, 3.0)

  def testAddDiscreteParamIntegerFeasibleValues(self):
    """Test a Discrete parameter with integer feasible values."""
    space = base_study_config.SearchSpace()
    self.assertEmpty(space.parameters)
    _ = space.select_root().add_discrete_param(
        'd1', [101, 15.0, 21.0], default_value=15.0)
    self.assertLen(space.parameters, 1)
    self.assertEqual(space.parameters[0].name, 'd1')
    self.assertEqual(space.parameters[0].type, pc.ParameterType.DISCRETE)
    self.assertEqual(space.parameters[0].bounds, (15.0, 101.0))
    self.assertEqual(space.parameters[0].scale_type, pc.ScaleType.LINEAR)
    self.assertEmpty(space.parameters[0].matching_parent_values)
    self.assertEmpty(space.parameters[0].child_parameter_configs)
    self.assertEqual(space.parameters[0].feasible_values, [15.0, 21.0, 101])
    self.assertEqual(space.parameters[0].default_value, 15.0)
    self.assertEqual(space.parameters[0].external_type, pc.ExternalType.INTEGER)

  def testAddDiscreteParamFloatFeasibleValues(self):
    """Test a Discrete parameter with float feasible values."""
    space = base_study_config.SearchSpace()
    self.assertEmpty(space.parameters)
    _ = space.select_root().add_discrete_param(
        'd1', [15.1, 21.0, 101], default_value=15.1)
    self.assertLen(space.parameters, 1)
    self.assertEqual(space.parameters[0].external_type, pc.ExternalType.FLOAT)

  def testAddBooleanParam(self):
    """Test a Boolean parameter."""
    space = base_study_config.SearchSpace()
    self.assertEmpty(space.parameters)
    _ = space.select_root().add_bool_param('b1', default_value=True)
    self.assertLen(space.parameters, 1)
    self.assertEqual(space.parameters[0].name, 'b1')
    self.assertEqual(space.parameters[0].type, pc.ParameterType.CATEGORICAL)
    with self.assertRaisesRegex(ValueError,
                                'Accessing bounds of a categorical.*'):
      _ = space.parameters[0].bounds
    self.assertIsNone(space.parameters[0].scale_type)
    self.assertEmpty(space.parameters[0].matching_parent_values)
    self.assertEmpty(space.parameters[0].child_parameter_configs)
    self.assertEqual(space.parameters[0].feasible_values, ['False', 'True'])
    self.assertEqual(space.parameters[0].default_value, 'True')
    self.assertEqual(space.parameters[0].external_type, pc.ExternalType.BOOLEAN)

  def testAddBooleanParamWithFalseDefault(self):
    """Test a Boolean parameter."""
    space = base_study_config.SearchSpace()
    self.assertEmpty(space.parameters)
    _ = space.select_root().add_bool_param('b1', default_value=False)
    self.assertLen(space.parameters, 1)
    self.assertEqual(space.parameters[0].default_value, 'False')

  def testAddTwoFloatParams(self):
    space = base_study_config.SearchSpace()
    self.assertEmpty(space.parameters)
    _ = space.select_root().add_float_param(
        'f1', 1.0, 15.0, default_value=3.0, scale_type=pc.ScaleType.LOG)
    _ = space.select_root().add_float_param(
        'f2', 2.0, 16.0, default_value=4.0, scale_type=pc.ScaleType.REVERSE_LOG)

    self.assertLen(space.parameters, 2)

    self.assertEqual(space.parameters[0].name, 'f1')
    self.assertEqual(space.parameters[0].type, pc.ParameterType.DOUBLE)
    self.assertEqual(space.parameters[0].bounds, (1.0, 15.0))
    self.assertEqual(space.parameters[0].scale_type, pc.ScaleType.LOG)
    self.assertEmpty(space.parameters[0].matching_parent_values)
    self.assertEmpty(space.parameters[0].child_parameter_configs)
    with self.assertRaisesRegex(ValueError, 'feasible_values is invalid.*'):
      _ = space.parameters[0].feasible_values
    self.assertEqual(space.parameters[0].default_value, 3.0)

    self.assertEqual(space.parameters[1].name, 'f2')
    self.assertEqual(space.parameters[1].type, pc.ParameterType.DOUBLE)
    self.assertEqual(space.parameters[1].bounds, (2.0, 16.0))
    self.assertEqual(space.parameters[1].scale_type, pc.ScaleType.REVERSE_LOG)
    self.assertEmpty(space.parameters[1].matching_parent_values)
    self.assertEmpty(space.parameters[1].child_parameter_configs)
    with self.assertRaisesRegex(ValueError, 'feasible_values is invalid.*'):
      _ = space.parameters[1].feasible_values
    self.assertEqual(space.parameters[1].default_value, 4.0)

  def testChainAddTwoFloatParams(self):
    space = base_study_config.SearchSpace()
    self.assertEmpty(space.parameters)
    root = space.select_root()
    root.add_float_param(
        'f1', 1.0, 15.0, default_value=3.0, scale_type=pc.ScaleType.LOG)
    root.add_float_param(
        'f2', 2.0, 16.0, default_value=4.0, scale_type=pc.ScaleType.REVERSE_LOG)

    self.assertLen(space.parameters, 2)

    self.assertEqual(space.parameters[0].name, 'f1')
    self.assertEqual(space.parameters[0].type, pc.ParameterType.DOUBLE)
    self.assertEqual(space.parameters[0].bounds, (1.0, 15.0))
    self.assertEqual(space.parameters[0].scale_type, pc.ScaleType.LOG)
    self.assertEmpty(space.parameters[0].matching_parent_values)
    self.assertEmpty(space.parameters[0].child_parameter_configs)
    with self.assertRaisesRegex(ValueError, 'feasible_values is invalid.*'):
      _ = space.parameters[0].feasible_values
    self.assertEqual(space.parameters[0].default_value, 3.0)

    self.assertEqual(space.parameters[1].name, 'f2')
    self.assertEqual(space.parameters[1].type, pc.ParameterType.DOUBLE)
    self.assertEqual(space.parameters[1].bounds, (2.0, 16.0))
    self.assertEqual(space.parameters[1].scale_type, pc.ScaleType.REVERSE_LOG)
    self.assertEmpty(space.parameters[1].matching_parent_values)
    self.assertEmpty(space.parameters[1].child_parameter_configs)
    with self.assertRaisesRegex(ValueError, 'feasible_values is invalid.*'):
      _ = space.parameters[1].feasible_values
    self.assertEqual(space.parameters[1].default_value, 4.0)

  def testMultidimensionalParameters(self):
    space = base_study_config.SearchSpace()
    self.assertEmpty(space.parameters)
    selector0 = space.select_root().add_float_param(
        'f', 1.0, 15.0, default_value=3.0, scale_type=pc.ScaleType.LOG, index=0)
    selector1 = space.select_root().add_float_param(
        'f',
        2.0,
        10.0,
        default_value=4.0,
        scale_type=pc.ScaleType.LINEAR,
        index=1)
    # Test the returned selectors.
    self.assertEqual(selector0.path_string, '')
    self.assertEqual(selector0.parameter_name, 'f[0]')
    self.assertEqual(selector0.parameter_values, [])
    self.assertEqual(selector1.path_string, '')
    self.assertEqual(selector1.parameter_name, 'f[1]')
    self.assertEqual(selector1.parameter_values, [])
    # Test the search space.
    self.assertLen(space.parameters, 2)
    self.assertEqual(space.parameters[0].name, 'f[0]')
    self.assertEqual(space.parameters[0].type, pc.ParameterType.DOUBLE)
    self.assertEqual(space.parameters[0].bounds, (1.0, 15.0))
    self.assertEqual(space.parameters[0].scale_type, pc.ScaleType.LOG)
    self.assertEmpty(space.parameters[0].matching_parent_values)
    self.assertEmpty(space.parameters[0].child_parameter_configs)
    with self.assertRaisesRegex(ValueError, 'feasible_values is invalid.*'):
      _ = space.parameters[0].feasible_values
    self.assertEqual(space.parameters[0].default_value, 3.0)

    self.assertEqual(space.parameters[1].name, 'f[1]')
    self.assertEqual(space.parameters[1].type, pc.ParameterType.DOUBLE)
    self.assertEqual(space.parameters[1].bounds, (2.0, 10.0))
    self.assertEqual(space.parameters[1].scale_type, pc.ScaleType.LINEAR)
    self.assertEmpty(space.parameters[1].matching_parent_values)
    self.assertEmpty(space.parameters[1].child_parameter_configs)
    with self.assertRaisesRegex(ValueError, 'feasible_values is invalid.*'):
      _ = space.parameters[1].feasible_values
    self.assertEqual(space.parameters[1].default_value, 4.0)

  def testConditionalParameters(self):
    space = base_study_config.SearchSpace()
    self.assertEmpty(space.parameters)
    root = space.select_root()
    root.add_categorical_param(
        'model_type', ['linear', 'dnn'], default_value='dnn')
    # Test the selector.
    self.assertEqual(root.path_string, '')
    self.assertEqual(root.parameter_name, '')
    self.assertEqual(root.parameter_values, [])
    # Test the search space.
    self.assertLen(space.parameters, 1)
    self.assertEqual(space.parameters[0].name, 'model_type')
    self.assertEqual(space.parameters[0].type, pc.ParameterType.CATEGORICAL)
    with self.assertRaisesRegex(ValueError,
                                'Accessing bounds of a categorical.*'):
      _ = space.parameters[0].bounds
    self.assertIsNone(space.parameters[0].scale_type)
    self.assertEmpty(space.parameters[0].matching_parent_values)
    self.assertEmpty(space.parameters[0].child_parameter_configs)
    self.assertEqual(space.parameters[0].feasible_values, ['dnn', 'linear'])
    self.assertEqual(space.parameters[0].default_value, 'dnn')

    dnn = root.select('model_type', ['dnn'])
    # Test the selector.
    self.assertEqual(dnn.path_string, '')
    self.assertEqual(dnn.parameter_name, 'model_type')
    self.assertEqual(dnn.parameter_values, ['dnn'])
    dnn.add_float_param(
        'learning_rate',
        0.0001,
        1.0,
        default_value=0.001,
        scale_type=base_study_config.ScaleType.LOG)
    # Test the search space.
    self.assertLen(space.parameters, 1)

    linear = root.select('model_type', ['linear'])
    # Test the selector.
    self.assertEqual(linear.path_string, '')
    self.assertEqual(linear.parameter_name, 'model_type')
    self.assertEqual(linear.parameter_values, ['linear'])
    linear.add_float_param(
        'learning_rate',
        0.1,
        1.0,
        default_value=0.1,
        scale_type=base_study_config.ScaleType.LOG)
    # Test the search space.
    self.assertLen(space.parameters, 1)

    dnn_optimizer = dnn.add_categorical_param('optimizer_type',
                                              ['adam', 'adagrad'])
    # Test the search space.
    self.assertLen(space.parameters, 1)
    # Test the selector.
    self.assertEqual(dnn_optimizer.path_string, 'model_type=dnn')
    self.assertEqual(dnn_optimizer.parameter_name, 'optimizer_type')
    self.assertEqual(dnn_optimizer.parameter_values, [])

    # Chained select() calls, path length of 1.
    lr = root.select('model_type', ['dnn']).select(
        'optimizer_type', ['adam']).add_float_param(
            'learning_rate',
            0.1,
            1.0,
            default_value=0.1,
            scale_type=base_study_config.ScaleType.LOG)
    # Test the search space.
    self.assertLen(space.parameters, 1)
    # Test the selector.
    self.assertEqual(lr.parameter_name, 'learning_rate')
    self.assertEqual(lr.parameter_values, [])
    self.assertEqual(lr.path_string, 'model_type=dnn/optimizer_type=adam')

    # Chained select() calls, path length of 2.
    ko = root.select('model_type', ['dnn']).select('optimizer_type',
                                                   ['adam']).add_bool_param(
                                                       'use_keras_optimizer',
                                                       default_value=False)
    # Test the search space.
    self.assertLen(space.parameters, 1)
    # Test the selector.
    self.assertEqual(ko.parameter_name, 'use_keras_optimizer')
    self.assertEqual(ko.parameter_values, [])
    self.assertEqual(ko.path_string, 'model_type=dnn/optimizer_type=adam')

    ko.select_values(['True'])
    self.assertEqual(ko.parameter_values, ['True'])

    selector = ko.add_float_param('keras specific', 1.3, 2.4, default_value=2.1)
    # Test the search space.
    self.assertLen(space.parameters, 1)
    # Test the selector.
    self.assertEqual(selector.parameter_name, 'keras specific')
    self.assertEqual(selector.parameter_values, [])
    self.assertEqual(
        selector.path_string,
        'model_type=dnn/optimizer_type=adam/use_keras_optimizer=True')

    # Selects more than one node.
    # selectors = dnn.select_all('optimizer_type', ['adam'])
    # self.assertLen(selectors, 2)

  def testConditionalParametersWithReturnedSelectors(self):
    space = base_study_config.SearchSpace()
    self.assertEmpty(space.parameters)
    root = space.select_root()
    model_type = root.add_categorical_param('model_type', ['linear', 'dnn'])
    learning_rate = model_type.select_values(['dnn']).add_float_param(
        'learning_rate',
        0.1,
        1.0,
        default_value=0.001,
        scale_type=base_study_config.ScaleType.LOG)
    # Test the search space.
    self.assertLen(space.parameters, 1)
    # Test the selectors.
    self.assertEqual(model_type.parameter_values, ['dnn'])
    self.assertEqual(learning_rate.parameter_name, 'learning_rate')
    self.assertEqual(learning_rate.parameter_values, [])
    self.assertEqual(learning_rate.path_string, 'model_type=dnn')

    # It is possible to select different values for the same selector.
    optimizer_type = model_type.select_values(['linear',
                                               'dnn']).add_categorical_param(
                                                   'optimizer_type',
                                                   ['adam', 'adagrad'])
    # Test the search space.
    self.assertLen(space.parameters, 1)
    # Test the selectors.
    self.assertEqual(model_type.parameter_values, ['linear', 'dnn'])
    self.assertEqual(optimizer_type.parameter_name, 'optimizer_type')
    self.assertEqual(optimizer_type.parameter_values, [])
    self.assertEqual(optimizer_type.path_string, 'model_type=linear')

  @parameterized.named_parameters(
      ('Multi', 'units[0]', ('units', 0)),
      ('Multi2', 'with_underscore[1]', ('with_underscore', 1)),
      ('NotMulti', 'units', None),
      ('NotMulti2', 'with space', None),
      ('NotMulti3', 'with[8]space', None),
      ('NotMulti4', 'units[0][4]', ('units[0]', 4)),
      ('GinStyle', '_gin.ambient_net_exp_from_vec.block_type[3]',
       ('_gin.ambient_net_exp_from_vec.block_type', 3)),
  )
  def testParseMultiDimensionalParameterName(self, name, expected):
    base_name_index = base_study_config.SearchSpaceSelector.parse_multi_dimensional_parameter_name(
        name)
    self.assertEqual(base_name_index, expected)


class SearchSpaceContainsTest(absltest.TestCase):

  def _space(self):
    space = base_study_config.SearchSpace()
    root = space.select_root()
    root.add_float_param('learning-rate', 1e-4, 1e-2)
    root.add_categorical_param('optimizer', ['adagrad', 'adam', 'experimental'])
    return space

  def testFloatCat1(self):
    self._space().assert_contains(
        trial.ParameterDict({
            'optimizer': 'adagrad',
            'learning-rate': 1e-2
        }))

  def testFloatCat2(self):
    self.assertFalse(self._space().contains(
        trial.ParameterDict({
            'optimizer': 'adagrad',
            'BADPARAM': 1e-2
        })))

  def testFloatCat3(self):
    self.assertFalse(self._space().contains(
        trial.ParameterDict({
            'optimizer': 'adagrad',
            'learning-rate': 1e-2,
            'BADPARAM': 1e-2
        })))

  def testFloatCat4(self):
    self.assertFalse(self._space().contains(
        trial.ParameterDict({
            'optimizer': 'adagrad',
            'learning-rate': 1e2
        })))


if __name__ == '__main__':
  absltest.main()



"""OSS Vizier client."""

# TODO: Raise vizier-specific exceptions.

from typing import Iterable, Any, Collection, Mapping, Optional, Type

import attr
from vizier._src.pyvizier.client import client_abc
from vizier.service import pyvizier as vz
from vizier.service import vizier_client
from vizier.service import vizier_service_pb2_grpc


@attr.define
class _EnviromentVariables:
  service_endpoint: str = attr.field(
      default='UNSET', validator=attr.validators.instance_of(str))


environment_variables = _EnviromentVariables()

_UNUSED_CLIENT_ID = 'Unused client id.'


def _get_stub() -> vizier_service_pb2_grpc.VizierServiceStub:
  return vizier_client.create_server_stub(
      environment_variables.service_endpoint)


@attr.define
class Trial(client_abc.TrialInterface):
  """Trial class."""

  _client: vizier_client.VizierClient = attr.field()
  _id: int = attr.field(validator=attr.validators.instance_of(int))

  @property
  def id(self) -> int:
    return self._id

  @property
  def parameters(self) -> Mapping[str, Any]:
    trial = self.materialize(include_all_measurements=False)
    study_config = self._client.get_study_config()
    return study_config.trial_parameters(vz.TrialConverter.to_proto(trial))

  def delete(self) -> None:
    self._client.delete_trial(self._id)

  def complete(
      self,
      measurement: Optional[vz.Measurement] = None,
      *,
      infeasible_reason: Optional[str] = None) -> Optional[vz.Measurement]:
    self._trial = self._client.complete_trial(self._id, measurement,
                                              infeasible_reason)
    return self._trial.final_measurement

  def check_early_stopping(self) -> bool:
    return self._client.should_trial_stop(self._id)

  def add_measurement(self, measurement: vz.Measurement) -> None:
    self._client.report_intermediate_objective_value(
        int(measurement.steps),
        measurement.elapsed_secs,
        [{k: v.value for k, v in measurement.metrics.items()}],
        trial_id=self._id)

  def materialize(self, *, include_all_measurements: bool = True) -> vz.Trial:
    trial = self._client.get_trial(self._id)
    if not include_all_measurements:
      trial.measurements.clear()
    return trial


@attr.define
class Study(client_abc.StudyInterface):
  """Responsible for study-level operations."""
  _client: vizier_client.VizierClient = attr.field()

  @property
  def resource_name(self) -> str:
    return self._client.study_resource_name

  def _trial_client(self, trial_id: int) -> Trial:
    return Trial(self._client, trial_id)

  def suggest(self,
              *,
              count: Optional[int] = None,
              client_id: str = 'default_client_id') -> Collection[Trial]:
    return [
        self._trial_client(t.id) for t in self._client.get_suggestions(
            count, client_id_override=client_id)
    ]

  def delete(self) -> None:
    self._client.delete_study()

  def _add_trial(self, trial: vz.Trial) -> Trial:
    return self._trial_client(self._client.add_trial(trial).id)

  def trials(self,
             trial_filter: Optional[vz.TrialFilter] = None) -> Iterable[Trial]:
    all_trials = self._client.list_trials()
    trial_filter = trial_filter or vz.TrialFilter()
    for t in filter(trial_filter, all_trials):
      yield self._trial_client(t.id)

  def get_trial(self, trial_id: int, /) -> Trial:
    trial = self._trial_client(trial_id)
    try:
      # Check if the trial actually exists.
      trial.materialize(include_all_measurements=False)
      return trial
    except KeyError as err:
      raise client_abc.ResourceNotFoundError(
          f'Study f{self.resource_name} does not have '
          f'Trial {trial_id}.') from err

  def optimal_trials(self) -> list[Trial]:
    trials = self._client.list_optimal_trials()
    return [self._trial_client(t.id) for t in trials]

  def materialize_problem_statement(self) -> vz.ProblemStatement:
    return self._client.get_study_config().to_problem()

  @classmethod
  def from_resource_name(cls: Type['Study'], name: str, /) -> 'Study':
    client = vizier_client.VizierClient(_get_stub(), name, _UNUSED_CLIENT_ID)
    try:
      _ = client.get_study_config()  # Make sure study exists.
    except Exception as err:
      raise client_abc.ResourceNotFoundError() from err
    return Study(client)

  @classmethod
  def from_study_config(cls, config: vz.StudyConfig, /, *, owner: str,
                        study_id: str) -> 'Study':
    """Create study from StudyConfig.

    Args:
      config: OSS Study configuration. It is platform-specific. It is ignored if
        `owner` already has Study with `study_id`. # TODO: Instead
        of ignoring it, compare with the existing # study and return error if
        there's a change.
      owner: Owner of the study.
      study_id: Unique identifier within the same owner.

    Returns:
      Study.
    """
    return Study(
        vizier_client.create_or_load_study(
            environment_variables.service_endpoint,
            owner_id=owner,
            client_id=_UNUSED_CLIENT_ID,
            study_id=study_id,
            study_config=config))



"""StudyConfig used by pythia policies."""

import attr

from vizier._src.pyvizier.shared import base_study_config


@attr.define(frozen=True, init=True)
class StudyDescriptor:
  """Light-weight summary for Study."""

  config: base_study_config.ProblemStatement = attr.ib(
      init=True,
      validator=[
          attr.validators.optional(
              attr.validators.instance_of(base_study_config.ProblemStatement))
      ])

  guid: str = attr.ib(
      init=True,
      validator=[attr.validators.optional(attr.validators.instance_of(str))],
      kw_only=True)

  max_trial_id: int = attr.ib(
      init=True,
      validator=[attr.validators.optional(attr.validators.instance_of(int))],
      kw_only=True)



"""Converters for OSS Vizier's protos from/to PyVizier's classes."""
import datetime
import logging
from typing import Iterable, List, Optional, Sequence, Tuple, Union
from absl import logging

from vizier._src.pyvizier.oss import metadata_util
from vizier._src.pyvizier.shared import base_study_config
from vizier._src.pyvizier.shared import common
from vizier._src.pyvizier.shared import parameter_config
from vizier._src.pyvizier.shared import trial
from vizier.service import study_pb2

ScaleType = parameter_config.ScaleType
_ScaleTypePb2 = study_pb2.StudySpec.ParameterSpec.ScaleType
ParameterType = parameter_config.ParameterType
MonotypeParameterSequence = parameter_config.MonotypeParameterSequence


class _ScaleTypeMap:
  """Proto converter for scale type."""
  _pyvizier_to_proto = {
      parameter_config.ScaleType.LINEAR:
          _ScaleTypePb2.UNIT_LINEAR_SCALE,
      parameter_config.ScaleType.LOG:
          _ScaleTypePb2.UNIT_LOG_SCALE,
      parameter_config.ScaleType.REVERSE_LOG:
          _ScaleTypePb2.UNIT_REVERSE_LOG_SCALE,
  }
  _proto_to_pyvizier = {v: k for k, v in _pyvizier_to_proto.items()}

  @classmethod
  def to_proto(cls, pyvizier: parameter_config.ScaleType) -> _ScaleTypePb2:
    return cls._pyvizier_to_proto[pyvizier]

  @classmethod
  def from_proto(cls, proto: _ScaleTypePb2) -> parameter_config.ScaleType:
    return cls._proto_to_pyvizier[proto]


class ParameterConfigConverter:
  """Converter for ParameterConfig."""

  @classmethod
  def _set_bounds(cls, proto: study_pb2.StudySpec.ParameterSpec, lower: float,
                  upper: float, parameter_type: ParameterType):
    """Sets the proto's min_value and max_value fields."""
    if parameter_type == ParameterType.INTEGER:
      proto.integer_value_spec.min_value = lower
      proto.integer_value_spec.max_value = upper
    elif parameter_type == ParameterType.DOUBLE:
      proto.double_value_spec.min_value = lower
      proto.double_value_spec.max_value = upper

  @classmethod
  def _set_feasible_points(cls, proto: study_pb2.StudySpec.ParameterSpec,
                           feasible_points: Sequence[float]):
    """Sets the proto's feasible_points field."""
    feasible_points = sorted(feasible_points)
    proto.discrete_value_spec.ClearField('values')
    proto.discrete_value_spec.values.extend(feasible_points)

  @classmethod
  def _set_categories(cls, proto: study_pb2.StudySpec.ParameterSpec,
                      categories: Sequence[str]):
    """Sets the protos' categories field."""
    proto.categorical_value_spec.ClearField('values')
    proto.categorical_value_spec.values.extend(categories)

  @classmethod
  def _set_default_value(cls, proto: study_pb2.StudySpec.ParameterSpec,
                         default_value: Union[float, int, str]):
    """Sets the protos' default_value field."""
    which_pv_spec = proto.WhichOneof('parameter_value_spec')
    getattr(proto, which_pv_spec).default_value.value = default_value

  @classmethod
  def _matching_parent_values(
      cls, proto: study_pb2.StudySpec.ParameterSpec.ConditionalParameterSpec
  ) -> MonotypeParameterSequence:
    """Returns the matching parent values, if set."""
    oneof_name = proto.WhichOneof('parent_value_condition')
    if not oneof_name:
      return []
    if oneof_name in ('parent_discrete_values', 'parent_int_values',
                      'parent_categorical_values'):
      return list(getattr(getattr(proto, oneof_name), 'values'))
    raise ValueError('Unknown matching_parent_vals: {}'.format(oneof_name))

  @classmethod
  def from_proto(
      cls,
      proto: study_pb2.StudySpec.ParameterSpec,
      *,
      strict_validation: bool = False) -> parameter_config.ParameterConfig:
    """Creates a ParameterConfig.

    Args:
      proto:
      strict_validation: If True, raise ValueError to enforce that
        from_proto(proto).to_proto == proto.

    Returns:
      ParameterConfig object

    Raises:
      ValueError: See the "strict_validtion" arg documentation.
    """
    feasible_values = []
    oneof_name = proto.WhichOneof('parameter_value_spec')
    if oneof_name == 'integer_value_spec':
      bounds = (int(proto.integer_value_spec.min_value),
                int(proto.integer_value_spec.max_value))
    elif oneof_name == 'double_value_spec':
      bounds = (proto.double_value_spec.min_value,
                proto.double_value_spec.max_value)
    elif oneof_name == 'discrete_value_spec':
      bounds = None
      feasible_values = proto.discrete_value_spec.values
    elif oneof_name == 'categorical_value_spec':
      bounds = None
      feasible_values = proto.categorical_value_spec.values

    default_value = None
    if getattr(proto, oneof_name).default_value.value:
      default_value = getattr(proto, oneof_name).default_value.value

    if proto.conditional_parameter_specs:
      children = []
      for conditional_ps in proto.conditional_parameter_specs:
        parent_values = cls._matching_parent_values(conditional_ps)
        children.append(
            (parent_values, cls.from_proto(conditional_ps.parameter_spec)))
    else:
      children = None

    scale_type = None
    if proto.scale_type:
      scale_type = _ScaleTypeMap.from_proto(proto.scale_type)

    try:
      config = parameter_config.ParameterConfig.factory(
          name=proto.parameter_id,
          feasible_values=feasible_values,
          bounds=bounds,
          children=children,
          scale_type=scale_type,
          default_value=default_value)
    except ValueError as e:
      raise ValueError(
          'The provided proto was misconfigured. {}'.format(proto)) from e

    if strict_validation and cls.to_proto(config) != proto:
      raise ValueError(
          'The provided proto was misconfigured. Expected: {} Given: {}'.format(
              cls.to_proto(config), proto))
    return config

  @classmethod
  def _set_child_parameter_configs(
      cls, parent_proto: study_pb2.StudySpec.ParameterSpec,
      pc: parameter_config.ParameterConfig):
    """Sets the parent_proto's conditional_parameter_specs field.

    Args:
      parent_proto: Modified in place.
      pc: Parent ParameterConfig to copy children from.

    Raises:
      ValueError: If the child configs are invalid
    """
    children: List[Tuple[MonotypeParameterSequence,
                         parameter_config.ParameterConfig]] = []
    for child in pc.child_parameter_configs:
      children.append((child.matching_parent_values, child))
    if not children:
      return

    parent_proto.ClearField('conditional_parameter_specs')
    for child_pair in children:
      if len(child_pair) != 2:
        raise ValueError("""Each element in children must be a tuple of
            (Sequence of valid parent values,  ParameterConfig)""")

    logging.debug('_set_child_parameter_configs: parent_proto=%s, children=%s',
                  parent_proto, children)
    for unsorted_parent_values, child in children:
      parent_values = sorted(unsorted_parent_values)
      child_proto = cls.to_proto(child.clone_without_children)
      conditional_parameter_spec = study_pb2.StudySpec.ParameterSpec.ConditionalParameterSpec(
          parameter_spec=child_proto)

      if parent_proto.HasField('discrete_value_spec'):
        conditional_parameter_spec.parent_discrete_values.values[:] = parent_values
      elif parent_proto.HasField('categorical_value_spec'):
        conditional_parameter_spec.parent_categorical_values.values[:] = parent_values
      elif parent_proto.HasField('integer_value_spec'):
        conditional_parameter_spec.parent_int_values.values[:] = parent_values
      else:
        raise ValueError('DOUBLE type cannot have child parameters')
      if child.child_parameter_configs:
        cls._set_child_parameter_configs(child_proto, child)
      parent_proto.conditional_parameter_specs.extend(
          [conditional_parameter_spec])

  @classmethod
  def to_proto(
      cls, pc: parameter_config.ParameterConfig
  ) -> study_pb2.StudySpec.ParameterSpec:
    """Returns a ParameterConfig Proto."""
    proto = study_pb2.StudySpec.ParameterSpec(parameter_id=pc.name)
    if pc.type == ParameterType.DISCRETE:
      cls._set_feasible_points(proto, [float(v) for v in pc.feasible_values])
    elif pc.type == ParameterType.CATEGORICAL:
      cls._set_categories(proto, pc.feasible_values)
    elif pc.type in (ParameterType.INTEGER, ParameterType.DOUBLE):
      cls._set_bounds(proto, pc.bounds[0], pc.bounds[1], pc.type)
    else:
      raise ValueError('Invalid ParameterConfig: {}'.format(pc))
    if pc.scale_type is not None and pc.scale_type != ScaleType.UNIFORM_DISCRETE:
      proto.scale_type = _ScaleTypeMap.to_proto(pc.scale_type)
    if pc.default_value is not None:
      cls._set_default_value(proto, pc.default_value)

    cls._set_child_parameter_configs(proto, pc)
    return proto


class ParameterValueConverter:
  """Converter for trial.ParameterValue."""

  @classmethod
  def from_proto(
      cls, proto: study_pb2.Trial.Parameter) -> Optional[trial.ParameterValue]:
    """Returns whichever value that is populated, or None."""
    value_proto = proto.value
    oneof_name = value_proto.WhichOneof('kind')
    potential_value = getattr(value_proto, oneof_name)
    if isinstance(potential_value, float) or isinstance(
        potential_value, str) or isinstance(potential_value, bool):
      return trial.ParameterValue(potential_value)
    else:
      return None

  @classmethod
  def to_proto(cls, parameter_value: trial.ParameterValue,
               name: str) -> study_pb2.Trial.Parameter:
    """Returns Parameter Proto."""
    proto = study_pb2.Trial.Parameter(parameter_id=name)

    if isinstance(parameter_value.value, int):
      proto.value.number_value = parameter_value.value
    elif isinstance(parameter_value.value, bool):
      proto.value.bool_value = parameter_value.value
    elif isinstance(parameter_value.value, float):
      proto.value.number_value = parameter_value.value
    elif isinstance(parameter_value.value, str):
      proto.value.string_value = parameter_value.value

    return proto


class MeasurementConverter:
  """Converter for trial.MeasurementConverter."""

  @classmethod
  def from_proto(cls, proto: study_pb2.Measurement) -> trial.Measurement:
    """Creates a valid instance from proto.

    Args:
      proto: Measurement proto.

    Returns:
      A valid instance of Measurement object. Metrics with invalid values
      are automatically filtered out.
    """

    metrics = dict()

    for metric in proto.metrics:
      if metric.metric_id in metrics and metrics[
          metric.metric_id].value != metric.value:
        logging.log_first_n(
            logging.ERROR, 'Duplicate metric of name "%s".'
            'The newly found value %s will be used and '
            'the previously found value %s will be discarded.'
            'This always happens if the proto has an empty-named metric.', 5,
            metric.metric_id, metric.value, metrics[metric.metric_id].value)
      try:
        metrics[metric.metric_id] = trial.Metric(value=metric.value)
      except ValueError:
        pass
    return trial.Measurement(
        metrics=metrics,
        elapsed_secs=proto.elapsed_duration.seconds,
        steps=proto.step_count)

  @classmethod
  def to_proto(cls, measurement: trial.Measurement) -> study_pb2.Measurement:
    """Converts to Measurement proto."""
    proto = study_pb2.Measurement()
    for name, metric in measurement.metrics.items():
      proto.metrics.add(metric_id=name, value=metric.value)

    proto.step_count = measurement.steps
    int_seconds = int(measurement.elapsed_secs)
    proto.elapsed_duration.seconds = int_seconds
    proto.elapsed_duration.nanos = int(1e9 *
                                       (measurement.elapsed_secs - int_seconds))
    return proto


class MetricInformationConverter:
  """A converter to/from study_pb2.StudySpec.MetricInformation."""

  @classmethod
  def from_proto(
      cls, proto: study_pb2.StudySpec.MetricSpec
  ) -> base_study_config.MetricInformation:
    """Converts a MetricInformation proto to a MetricInformation object."""
    if proto.goal not in list(base_study_config.ObjectiveMetricGoal):
      raise ValueError('Unknown MetricInformation.goal: {}'.format(proto.goal))

    return base_study_config.MetricInformation(
        name=proto.metric_id,
        goal=proto.goal,
        safety_threshold=None,
        safety_std_threshold=None,
        min_value=None,
        max_value=None)

  @classmethod
  def to_proto(
      cls, obj: base_study_config.MetricInformation
  ) -> study_pb2.StudySpec.MetricSpec:
    """Returns this object as a proto."""
    return study_pb2.StudySpec.MetricSpec(
        metric_id=obj.name, goal=obj.goal.value)


class SearchSpaceConverter:
  """A wrapper for study_pb2.StudySpec."""

  @classmethod
  def from_proto(cls,
                 proto: study_pb2.StudySpec) -> base_study_config.SearchSpace:
    """Extracts a SearchSpace object from a StudyConfig proto."""
    parameter_configs = []
    for pc in proto.parameters:
      parameter_configs.append(ParameterConfigConverter.from_proto(pc))
    # TODO: Remove _factory implementation.
    return base_study_config.SearchSpace._factory(  # pylint:disable=protected-access
        parameter_configs=parameter_configs)

  @classmethod
  def parameter_protos(
      cls, obj: base_study_config.SearchSpace
  ) -> List[study_pb2.StudySpec.ParameterSpec]:
    """Returns the search space as a List of ParameterConfig protos."""
    return [
        ParameterConfigConverter.to_proto(pc) for pc in obj._parameter_configs  # pylint:disable=protected-access
    ]


class MetricsConfigConverter:
  """A wrapper for study_pb2.StudySpec.MetricSpec's."""

  @classmethod
  def from_proto(
      cls, protos: Iterable[study_pb2.StudySpec.MetricSpec]
  ) -> base_study_config.MetricsConfig:
    return base_study_config.MetricsConfig(
        [MetricInformationConverter.from_proto(m) for m in protos])

  @classmethod
  def to_proto(
      cls, obj: base_study_config.MetricsConfig
  ) -> List[study_pb2.StudySpec.MetricSpec]:
    return [
        MetricInformationConverter.to_proto(metric) for metric in obj._metrics  # pylint:disable=protected-access
    ]


def _to_pyvizier_trial_status(
    proto_state: study_pb2.Trial.State) -> trial.TrialStatus:
  """from_proto conversion for Trial statuses."""
  if proto_state == study_pb2.Trial.State.REQUESTED:
    return trial.TrialStatus.REQUESTED
  elif proto_state == study_pb2.Trial.State.ACTIVE:
    return trial.TrialStatus.ACTIVE
  if proto_state == study_pb2.Trial.State.STOPPING:
    return trial.TrialStatus.STOPPING
  if proto_state == study_pb2.Trial.State.SUCCEEDED:
    return trial.TrialStatus.COMPLETED
  elif proto_state == study_pb2.Trial.State.INFEASIBLE:
    return trial.TrialStatus.COMPLETED
  else:
    return trial.TrialStatus.UNKNOWN


def _from_pyvizier_trial_status(status: trial.TrialStatus,
                                infeasible: bool) -> study_pb2.Trial.State:
  """to_proto conversion for Trial states."""
  if status == trial.TrialStatus.REQUESTED:
    return study_pb2.Trial.State.REQUESTED
  elif status == trial.TrialStatus.ACTIVE:
    return study_pb2.Trial.State.ACTIVE
  elif status == trial.TrialStatus.STOPPING:
    return study_pb2.Trial.State.STOPPING
  elif status == trial.TrialStatus.COMPLETED:
    if infeasible:
      return study_pb2.Trial.State.INFEASIBLE
    else:
      return study_pb2.Trial.State.SUCCEEDED
  else:
    return study_pb2.Trial.State.STATE_UNSPECIFIED


class TrialConverter:
  """Converter for trial.TrialConverter."""

  @classmethod
  def from_proto(cls, proto: study_pb2.Trial) -> trial.Trial:
    """Converts from Trial proto to object.

    Args:
      proto: Trial proto.

    Returns:
      A Trial object.
    """
    parameters = {}
    for parameter in proto.parameters:
      value = ParameterValueConverter.from_proto(parameter)
      if value is not None:
        if parameter.parameter_id in parameters:
          raise ValueError('Invalid trial proto contains duplicate parameter {}'
                           ': {}'.format(parameter.parameter_id, proto))
        parameters[parameter.parameter_id] = value
      else:
        logging.warning('A parameter without a value will be dropped: %s',
                        parameter)

    final_measurement = None
    if proto.HasField('final_measurement'):
      final_measurement = MeasurementConverter.from_proto(
          proto.final_measurement)

    completion_time = None
    infeasibility_reason = None
    if proto.state == study_pb2.Trial.State.SUCCEEDED:
      if proto.HasField('end_time'):
        completion_ts = proto.end_time.seconds + 1e-9 * proto.end_time.nanos
        completion_time = datetime.datetime.fromtimestamp(completion_ts)
    elif proto.state == study_pb2.Trial.State.INFEASIBLE:
      infeasibility_reason = proto.infeasible_reason

    metadata = trial.Metadata()
    for kv in proto.metadata:
      metadata.abs_ns(common.Namespace.decode(kv.ns))[kv.key] = (
          kv.proto if kv.HasField('proto') else kv.value)

    measurements = []
    for measure in proto.measurements:
      measurements.append(MeasurementConverter.from_proto(measure))

    creation_time = None
    if proto.HasField('start_time'):
      creation_ts = proto.start_time.seconds + 1e-9 * proto.start_time.nanos
      creation_time = datetime.datetime.fromtimestamp(creation_ts)
    return trial.Trial(
        id=int(proto.id),
        description=proto.name,
        assigned_worker=proto.client_id or None,
        is_requested=proto.state == proto.REQUESTED,
        stopping_reason=('stopping reason not supported yet'
                         if proto.state == proto.STOPPING else None),
        parameters=parameters,
        creation_time=creation_time,
        completion_time=completion_time,
        infeasibility_reason=infeasibility_reason,
        final_measurement=final_measurement,
        measurements=measurements,
        metadata=metadata)  # pytype: disable=wrong-arg-types

  @classmethod
  def from_protos(cls, protos: Sequence[study_pb2.Trial]) -> List[trial.Trial]:
    """Convenience wrapper for from_proto."""
    return [TrialConverter.from_proto(proto) for proto in protos]

  @classmethod
  def to_protos(cls, pytrials: Sequence[trial.Trial]) -> List[study_pb2.Trial]:
    return [TrialConverter.to_proto(pytrial) for pytrial in pytrials]

  @classmethod
  def to_proto(cls, pytrial: trial.Trial) -> study_pb2.Trial:
    """Converts a pyvizier Trial to a Trial proto."""
    proto = study_pb2.Trial()
    if pytrial.description is not None:
      proto.name = pytrial.description
    proto.id = str(pytrial.id)
    proto.state = _from_pyvizier_trial_status(pytrial.status,
                                              pytrial.infeasible)
    proto.client_id = pytrial.assigned_worker or ''

    for name, value in pytrial.parameters.items():
      proto.parameters.append(ParameterValueConverter.to_proto(value, name))

    # pytrial always adds an empty metric. Ideally, we should remove it if the
    # metric does not exist in the study config.
    if pytrial.final_measurement is not None:
      proto.final_measurement.CopyFrom(
          MeasurementConverter.to_proto(pytrial.final_measurement))

    for measurement in pytrial.measurements:
      proto.measurements.append(MeasurementConverter.to_proto(measurement))

    if pytrial.creation_time is not None:
      creation_secs = datetime.datetime.timestamp(pytrial.creation_time)
      proto.start_time.seconds = int(creation_secs)
      proto.start_time.nanos = int(1e9 * (creation_secs - int(creation_secs)))
    if pytrial.completion_time is not None:
      completion_secs = datetime.datetime.timestamp(pytrial.completion_time)
      proto.end_time.seconds = int(completion_secs)
      proto.end_time.nanos = int(1e9 * (completion_secs - int(completion_secs)))
    if pytrial.infeasibility_reason is not None:
      proto.infeasible_reason = pytrial.infeasibility_reason
    if pytrial.metadata is not None:
      for ns in pytrial.metadata.namespaces():
        ns_string = ns.encode()
        ns_layer = pytrial.metadata.abs_ns(ns)
        for key, value in ns_layer.items():
          metadata_util.assign(proto, key=key, ns=ns_string, value=value)
    return proto



"""Tests for vizier.pyvizier.shared.parameter_config."""

from typing import Any

from vizier._src.pyvizier.shared import parameter_config as pc
from absl.testing import absltest
from absl.testing import parameterized


class ParameterConfigFactoryTest(parameterized.TestCase):

  def testCreatesDoubleConfig(self):
    parameter_config = pc.ParameterConfig.factory(
        'name',
        bounds=(-1.0, 1.0),
        scale_type=pc.ScaleType.LINEAR,
        default_value=0.1)
    self.assertEqual(parameter_config.name, 'name')
    self.assertEqual(parameter_config.type, pc.ParameterType.DOUBLE)
    self.assertEqual(parameter_config.bounds, (-1, 1))
    self.assertEqual(parameter_config.scale_type, pc.ScaleType.LINEAR)
    self.assertEqual(parameter_config.default_value, 0.1)
    self.assertIsInstance(parameter_config.default_value, float)
    with self.assertRaises(ValueError):
      _ = parameter_config.feasible_values

    self.assertEqual(parameter_config.continuify(), parameter_config)

  def testCreatesIntegerConfig(self):
    parameter_config = pc.ParameterConfig.factory(
        'name', bounds=(1, 3), scale_type=pc.ScaleType.LOG, default_value=1)
    self.assertEqual(parameter_config.name, 'name')
    self.assertEqual(parameter_config.type, pc.ParameterType.INTEGER)
    self.assertEqual(parameter_config.feasible_values, [1, 2, 3])
    self.assertEqual(parameter_config.bounds, (1, 3))
    self.assertEqual(parameter_config.scale_type, pc.ScaleType.LOG)
    self.assertEqual(parameter_config.default_value, 1)
    self.assertIsInstance(parameter_config.default_value, int)

    self.assertEqual(
        parameter_config.continuify(),
        pc.ParameterConfig.factory(
            'name',
            bounds=(1.0, 3.0),
            scale_type=pc.ScaleType.LOG,
            default_value=1.0))

  def testCreatesDiscreteConfig(self):
    feasible_values = (-1, 3, 2)
    parameter_config = pc.ParameterConfig.factory(
        'name',
        feasible_values=feasible_values,
        scale_type=pc.ScaleType.UNIFORM_DISCRETE,
        default_value=2,
        external_type=pc.ExternalType.INTEGER)
    self.assertEqual(parameter_config.name, 'name')
    self.assertEqual(parameter_config.type, pc.ParameterType.DISCRETE)
    self.assertEqual(parameter_config.feasible_values, [-1, 2, 3])
    self.assertEqual(parameter_config.bounds, (-1, 3))
    self.assertEqual(parameter_config.scale_type, pc.ScaleType.UNIFORM_DISCRETE)
    self.assertEqual(parameter_config.default_value, 2)
    self.assertIsInstance(parameter_config.default_value, float)
    self.assertEqual(parameter_config.external_type, pc.ExternalType.INTEGER)

    self.assertEqual(
        parameter_config.continuify(),
        pc.ParameterConfig.factory(
            'name', bounds=(-1.0, 3.0), default_value=2.0))

  def testCreatesCategoricalConfig(self):
    feasible_values = ('b', 'a', 'c')
    parameter_config = pc.ParameterConfig.factory(
        'name', feasible_values=feasible_values, default_value='c')
    self.assertEqual(parameter_config.name, 'name')
    self.assertEqual(parameter_config.feasible_values, ['a', 'b', 'c'])
    self.assertEqual(parameter_config.default_value, 'c')
    with self.assertRaises(ValueError):
      _ = parameter_config.bounds

  def testCreatesDoubleConfigIntDefault(self):
    parameter_config = pc.ParameterConfig.factory(
        'name',
        bounds=(-1.0, 1.0),
        scale_type=pc.ScaleType.LINEAR,
        default_value=1)
    self.assertEqual(parameter_config.default_value, 1.0)
    self.assertIsInstance(parameter_config.default_value, float)

  def testCreatesDiscreteConfigDoubleDefault(self):
    feasible_values = (-1, 3, 2)
    parameter_config = pc.ParameterConfig.factory(
        'name',
        feasible_values=feasible_values,
        scale_type=pc.ScaleType.UNIFORM_DISCRETE,
        default_value=2.0)
    self.assertEqual(parameter_config.default_value, 2.0)
    self.assertIsInstance(parameter_config.default_value, float)

  def testCreatesIntegerConfigDoubleDefault(self):
    parameter_config = pc.ParameterConfig.factory(
        'name', bounds=(1, 3), scale_type=pc.ScaleType.LOG, default_value=2.0)
    self.assertEqual(parameter_config.default_value, 2.0)
    self.assertIsInstance(parameter_config.default_value, int)

  def testCreatesIntegerConfigInvalidDoubleDefault(self):
    with self.assertRaisesRegex(ValueError, 'default_value for an.*'):
      pc.ParameterConfig.factory(
          'name',
          bounds=(1, 3),
          scale_type=pc.ScaleType.LOG,
          default_value=2.0001)

  def testCreatesCategoricalConfigNoDefault(self):
    feasible_values = ('b', 'a', 'c')
    parameter_config = pc.ParameterConfig.factory(
        'name', feasible_values=feasible_values)
    self.assertIsNone(parameter_config.default_value)

  def testCreatesCategoricalConfigBadDefault(self):
    feasible_values = ('b', 'a', 'c')
    with self.assertRaisesRegex(ValueError,
                                'default_value has an incorrect type.*'):
      pc.ParameterConfig.factory(
          'name', feasible_values=feasible_values, default_value=0.1)

  def testRaisesErrorWhenNameIsEmpty(self):
    with self.assertRaises(ValueError):
      _ = pc.ParameterConfig.factory('', bounds=(-1.0, 1.0))

  def testRaisesErrorWhenOverSpecified(self):
    with self.assertRaises(ValueError):
      _ = pc.ParameterConfig.factory(
          'name', bounds=(-1.0, 1.0), feasible_values=['a', 'b', 'c'])

  @parameterized.named_parameters(
      ('HaveInfinity', (-float('inf'), 1)), ('HaveNan', (1, float('nan'))),
      ('HaveMixedTypes', (1, float(1))), ('AreWronglyOrdered', (1, -1)))
  def testRaisesErrorWhenBounds(self, bounds):
    with self.assertRaises(ValueError):
      _ = pc.ParameterConfig.factory('name', bounds=bounds)

  @parameterized.named_parameters(('HaveDuplicateCategories', ['a', 'a', 'b']),
                                  ('HaveDuplicateNumbers', [1.0, 2.0, 2.0]),
                                  ('HaveMixedTypes', ['a', 1, 2]))
  def testRaisesErrorWhenFeasibleValues(self, feasible_values):
    with self.assertRaises(ValueError):
      _ = pc.ParameterConfig.factory('name', feasible_values=feasible_values)


_child1 = pc.ParameterConfig.factory('double_child', bounds=(0.0, 1.0))
_child2 = pc.ParameterConfig.factory('integer_child', bounds=(0, 1))


class ParameterConfigFactoryTestWithChildren(parameterized.TestCase):

  @parameterized.named_parameters(
      ('IntParentValues', [([0], _child1), ([0, 1], _child2)]),
      ('FloatParentValues', [([0.0], _child1), ([0.0, 1.0], _child2)]))
  def testIntegerWithValid(self, children):
    p = pc.ParameterConfig.factory('parent', bounds=(0, 1), children=children)
    self.assertLen(p.child_parameter_configs, 2)
    self.assertEmpty(p.matching_parent_values)
    self.assertSameElements(p.child_parameter_configs[0].matching_parent_values,
                            children[0][0])
    self.assertSameElements(p.child_parameter_configs[1].matching_parent_values,
                            children[1][0])

  @parameterized.named_parameters(
      ('FloatParentValues', [([0.5], _child1)]),
      ('StringParentValues', [(['0'], _child1), (['0.0', '1.0'], _child2)]))
  def testIntegerWithInvalid(self, children):
    with self.assertRaises(TypeError):
      _ = pc.ParameterConfig.factory('parent', bounds=(0, 1), children=children)

  @parameterized.named_parameters(
      ('IntParentValues', [([0], _child1), ([1], _child2)]),
      ('FloatParentValues', [([0.0], _child1), ([0.0, 1.0], _child2)]))
  def testDiscreteWithValid(self, children):
    p = pc.ParameterConfig.factory(
        'parent', feasible_values=[0.0, 1.0], children=children)
    self.assertLen(p.child_parameter_configs, 2)
    self.assertEmpty(p.matching_parent_values)
    self.assertSameElements(p.child_parameter_configs[0].matching_parent_values,
                            children[0][0])
    self.assertSameElements(p.child_parameter_configs[1].matching_parent_values,
                            children[1][0])

  @parameterized.named_parameters(('StringParentValues', [(['0.0'], _child1),
                                                          (['0.0',
                                                            '1.0'], _child2)]))
  def testDiscreteWithInvalid(self, children):
    with self.assertRaises(TypeError):
      _ = pc.ParameterConfig.factory(
          'parent', feasible_values=[0.0, 1.0], children=children)

  @parameterized.named_parameters(  # pyformat: disable
      ('StringParentValues', [(['a'], _child1), (['a', 'b'], _child2)]))
  def testCategoricalWithValid(self, children):
    p = pc.ParameterConfig.factory(
        'parent', feasible_values=['a', 'b'], children=children)
    self.assertLen(p.child_parameter_configs, 2)
    self.assertEmpty(p.matching_parent_values)
    self.assertSameElements(p.child_parameter_configs[0].matching_parent_values,
                            children[0][0])
    self.assertSameElements(p.child_parameter_configs[1].matching_parent_values,
                            children[1][0])

  @parameterized.named_parameters(('StringParentValues', [(['0.0'], _child1),
                                                          (['1.0'], _child2)]))
  def testCategoricalWithInvalid(self, children):
    with self.assertRaises(TypeError):
      _ = pc.ParameterConfig.factory(
          'parent', feasible_values=[0.0, 1.0], children=children)

  def testAddChildren(self):
    children = [(['a'], _child1), (['a', 'b'], _child2)]
    p = pc.ParameterConfig.factory(
        'parent', feasible_values=['a', 'b'], children=children)
    new_children = [
        (['a'], pc.ParameterConfig.factory('double_child2', bounds=(1.0, 2.0))),
        (['b'],
         pc.ParameterConfig.factory(
             'categorical_child', feasible_values=['c', 'd'])),
    ]
    p2 = p.add_children(new_children)
    self.assertLen(p.child_parameter_configs, 2)
    self.assertSameElements([c.name for c in p.child_parameter_configs],
                            [c[1].name for c in children])

    self.assertLen(p2.child_parameter_configs, 4)
    expected_names = [c[1].name for c in children]
    expected_names += [c[1].name for c in new_children]
    got_names = [c.name for c in p2.child_parameter_configs]
    self.assertSameElements(got_names, expected_names)


class MergeTest(parameterized.TestCase):

  def test_merge_bounds(self):
    pc1 = pc.ParameterConfig.factory('pc1', bounds=(0.0, 2.0))
    pc2 = pc.ParameterConfig.factory('pc2', bounds=(-1.0, 1.0))
    self.assertEqual(
        pc.ParameterConfig.merge(pc1, pc2),
        pc.ParameterConfig.factory('pc1', bounds=(-1.0, 2.0)))

  def test_merge_discrete(self):
    pc1 = pc.ParameterConfig.factory(
        'pc1', feasible_values=[0.0, 2.0], scale_type=pc.ScaleType.LINEAR)
    pc2 = pc.ParameterConfig.factory('pc2', feasible_values=[-1.0, 0.0])
    self.assertEqual(
        pc.ParameterConfig.merge(pc1, pc2),
        pc.ParameterConfig.factory(
            'pc1',
            feasible_values=[-1.0, 0.0, 2.0],
            scale_type=pc.ScaleType.LINEAR))

  def test_merge_categorical(self):
    pc1 = pc.ParameterConfig.factory('pc1', feasible_values=['a', 'b'])
    pc2 = pc.ParameterConfig.factory('pc2', feasible_values=['a', 'c'])
    self.assertEqual(
        pc.ParameterConfig.merge(pc1, pc2),
        pc.ParameterConfig.factory('pc1', feasible_values=['a', 'b', 'c']))


class ParameterConfigContainsTest(parameterized.TestCase):

  @parameterized.parameters((1.0, True), (-2.0, False), (3.0, False))
  def testFloat(self, value: Any, expected: bool):
    config = pc.ParameterConfig.factory('pc1', bounds=(-1., 2.))
    self.assertEqual(config.contains(value), expected)

  @parameterized.parameters((1, True), (-2, False), (3, False), (1.5, False))
  def testInt(self, value: Any, expected: bool):
    config = pc.ParameterConfig.factory('pc1', bounds=(-1, 2))
    self.assertEqual(config.contains(value), expected)

  @parameterized.parameters((1.0, False), (2, True), (-1, True))
  def testDiscrete(self, value: Any, expected: bool):
    config = pc.ParameterConfig.factory('pc1', feasible_values=[-1., 0., 2.])
    self.assertEqual(config.contains(value), expected)

  @parameterized.parameters(('a', True), ('b', False), ('c', False))
  def testCategorical(self, value: Any, expected: bool):
    config = pc.ParameterConfig.factory(
        'pc1', feasible_values=['a', 'aa', 'aaa'])
    self.assertEqual(config.contains(value), expected)

  @parameterized.parameters((True, True), ('a', False), (0, False))
  def testBoolean(self, value: Any, expected: bool):
    config = pc.ParameterConfig.factory(
        'pc1', feasible_values=['true', 'false'])
    self.assertEqual(config.contains(value), expected)


class TraverseTest(parameterized.TestCase):

  @parameterized.named_parameters(('ShowChildrenTrue', True),
                                  ('ShowChildrenFalse', False))
  def testTraverse(self, show_children):
    grandchild1 = pc.ParameterConfig.factory('grandchild1', bounds=(-1.0, 1.0))
    grandchildren = [(['a'], grandchild1), (['b'], grandchild1)]
    child1 = pc.ParameterConfig.factory(
        'child1', feasible_values=['a', 'b'], children=grandchildren)

    child2 = pc.ParameterConfig.factory('child2', bounds=(0.0, 1.0))
    children = [([0], child1), ([1], child1), ([0, 1], child2)]
    parent = pc.ParameterConfig.factory(
        'parent', bounds=(0, 1), children=children)
    traversed_names = [
        pc.name for pc in parent.traverse(show_children=show_children)
    ]
    # Some parameter names are reused for separate child nodes, so they
    # will appear multiple times, but they are indeed separate parameters.
    self.assertEqual(traversed_names, [
        'parent', 'child1', 'grandchild1', 'grandchild1', 'child1',
        'grandchild1', 'grandchild1', 'child2'
    ])


if __name__ == '__main__':
  absltest.main()



"""Tests for vizier.pythia.policies.random_policy."""
from vizier import pythia
from vizier import pyvizier
from vizier._src.algorithms.policies import random_policy
from absl.testing import absltest


class RandomPolicyTest(absltest.TestCase):
  # TODO: Add conditional test case.

  def setUp(self):
    """Setups up search space."""
    self.study_config = pyvizier.StudyConfig()
    self.study_config.search_space.select_root().add_float_param(
        'double', min_value=-1.0, max_value=1.0)
    self.study_config.search_space.select_root().add_categorical_param(
        name='categorical', feasible_values=['a', 'b', 'c'])
    self.study_config.search_space.select_root().add_discrete_param(
        name='discrete', feasible_values=[0.1, 0.3, 0.5])
    self.study_config.search_space.select_root().add_int_param(
        name='int', min_value=1, max_value=5)

    self.policy_supporter = pythia.LocalPolicyRunner(self.study_config)
    self.policy = random_policy.RandomPolicy(
        policy_supporter=self.policy_supporter)
    super().setUp()

  def test_make_random_parameters(self):
    """Tests the internal random parameters function."""
    parameter_dict = random_policy.make_random_parameters(self.study_config)
    self.assertLen(parameter_dict, 4)

  def test_make_suggestions(self):
    """Tests random parameter generation wrapped around Policy."""
    num_suggestions = 5
    suggest_request = pythia.SuggestRequest(
        study_descriptor=self.policy_supporter.study_descriptor(),
        count=num_suggestions)

    suggestions = self.policy.suggest(suggest_request)
    self.assertLen(suggestions, num_suggestions)
    for suggestion in suggestions:
      self.assertLen(suggestion.parameters, 4)

  def test_make_early_stopping_decisions(self):
    """Checks if all ACTIVE/PENDING trials become completed in random order."""
    count = 10
    _ = self.policy_supporter.SuggestTrials(self.policy, count=10)

    request_trial_ids = [1, 2]
    trial_ids_stopped = set()
    for _ in range(count):
      request = pythia.EarlyStopRequest(
          study_descriptor=self.policy_supporter.study_descriptor(),
          trial_ids=request_trial_ids)
      early_stop_decisions = self.policy.early_stop(request)
      self.assertContainsSubset(
          request_trial_ids, [decision.id for decision in early_stop_decisions])

      for decision in early_stop_decisions:
        if decision.should_stop:
          # Stop trials that need to be stopped.
          self.policy_supporter.trials[decision.id - 1].complete(
              pyvizier.Measurement())
          trial_ids_stopped.add(decision.id)
    self.assertEqual(trial_ids_stopped, set(range(1, count + 1)))


if __name__ == '__main__':
  absltest.main()



from vizier import pyvizier
from vizier._src.pyvizier.multimetric import safety
from absl.testing import absltest


class SafeAcquisitionTest(absltest.TestCase):

  def testSafeMeasurements(self):
    info1 = pyvizier.MetricInformation(
        name='safety1',
        goal=pyvizier.ObjectiveMetricGoal.MAXIMIZE,
        safety_threshold=0.1)
    info2 = pyvizier.MetricInformation(
        name='safety2',
        goal=pyvizier.ObjectiveMetricGoal.MINIMIZE,
        safety_threshold=0.5)
    checker = safety.SafetyChecker(pyvizier.MetricsConfig([info1, info2]))
    safe_measurement = pyvizier.Measurement(metrics={
        'safety1': pyvizier.Metric(value=3.1),
        'safety2': pyvizier.Metric(value=0.3)
    })
    unsafe1_measurement = pyvizier.Measurement(metrics={
        'safety1': pyvizier.Metric(value=-3.1),
        'safety2': pyvizier.Metric(value=0.4)
    })
    unsafe2_measurement = pyvizier.Measurement(metrics={
        'safety1': pyvizier.Metric(value=3.1),
        'safety2': pyvizier.Metric(value=0.7)
    })
    no_safety_measurement = pyvizier.Measurement(
        metrics={'safety1': pyvizier.Metric(value=3.1)})

    self.assertListEqual(
        checker.are_measurements_safe([
            safe_measurement, unsafe1_measurement, unsafe2_measurement,
            no_safety_measurement
        ]), [True, False, False, True])

  def testSafeTrials(self):
    info1 = pyvizier.MetricInformation(
        name='obj', goal=pyvizier.ObjectiveMetricGoal.MAXIMIZE)
    info2 = pyvizier.MetricInformation(
        name='safety',
        goal=pyvizier.ObjectiveMetricGoal.MINIMIZE,
        safety_threshold=0.5)
    checker = safety.SafetyChecker(pyvizier.MetricsConfig([info1, info2]))

    safe_trial = pyvizier.Trial(
        final_measurement=pyvizier.Measurement(metrics={
            'obj': 0.2,
            'safety': 0.3
        }))
    unsafe_trial = pyvizier.Trial(
        final_measurement=pyvizier.Measurement(metrics={
            'obj': 0.2,
            'safety': 0.7
        }))
    no_safety_trial = pyvizier.Trial(
        final_measurement=pyvizier.Measurement(metrics={'obj': 0.2}))

    # Empty Trial is assumed to be safe.
    self.assertListEqual(
        checker.are_trials_safe(
            [safe_trial, unsafe_trial, no_safety_trial,
             pyvizier.Trial()]), [True, False, True, True])

  def testNoSafety(self):
    info1 = pyvizier.MetricInformation(
        name='obj', goal=pyvizier.ObjectiveMetricGoal.MAXIMIZE)
    checker = safety.SafetyChecker(pyvizier.MetricsConfig([info1]))

    safe_trial = pyvizier.Trial(
        final_measurement=pyvizier.Measurement(metrics={
            'obj': 0.2,
            'safety': 0.3
        }))
    unsafe_trial = pyvizier.Trial(
        final_measurement=pyvizier.Measurement(metrics={
            'obj': 0.2,
            'safety': 0.7
        }))

    # With no safety metrics, all Trials are safe..
    self.assertListEqual(
        checker.are_trials_safe([safe_trial, unsafe_trial,
                                 pyvizier.Trial()]), [True, True, True])


if __name__ == '__main__':
  absltest.main()



"""Import target for converters."""

from vizier.pyvizier.converters.core import DefaultModelInputConverter
from vizier.pyvizier.converters.core import DefaultModelOutputConverter
from vizier.pyvizier.converters.core import DefaultTrialConverter
from vizier.pyvizier.converters.core import dict_to_array
from vizier.pyvizier.converters.core import DictOf2DArrays
from vizier.pyvizier.converters.core import ModelInputConverter
from vizier.pyvizier.converters.core import ModelOutputConverter
from vizier.pyvizier.converters.core import NumpyArraySpec
from vizier.pyvizier.converters.core import NumpyArraySpecType
from vizier.pyvizier.converters.core import STUDY_ID_FIELD
from vizier.pyvizier.converters.core import TrialToArrayConverter
from vizier.pyvizier.converters.core import TrialToNumpyDict
from vizier.pyvizier.converters.spatio_temporal import DenseSpatioTemporalConverter
from vizier.pyvizier.converters.spatio_temporal import SparseSpatioTemporalConverter
from vizier.pyvizier.converters.spatio_temporal import TimedLabels
from vizier.pyvizier.converters.spatio_temporal import TimedLabelsExtractor



"""Vizier Service with in-RAM data storage."""

from concurrent import futures
import datetime
import time

import attr
import grpc
import portpicker
from vizier.service import datastore
from vizier.service import vizier_client
from vizier.service import vizier_server
from vizier.service import vizier_service_pb2_grpc


@attr.define
class LocalVizierTestService:
  """Local Vizier service with InRAM data storage."""
  _early_stop_recycle_period: datetime.timedelta = attr.field(
      init=False, default=datetime.timedelta(seconds=1))
  _port = attr.field(init=False, factory=portpicker.pick_unused_port)
  _servicer: vizier_server.VizierService = attr.field(init=False)
  _server: grpc.Server = attr.field(init=False)
  stub: vizier_service_pb2_grpc.VizierServiceStub = attr.field(init=False)

  @property
  def datastore(self) -> datastore.DataStore:
    return self._servicer.datastore

  @property
  def endpoint(self) -> str:
    return f'localhost:{self._port}'

  def __init__(self):
    self.__attrs_init__()
    self._servicer = vizier_server.VizierService(
        early_stop_recycle_period=self._early_stop_recycle_period)
    # Setup server.
    self._server = grpc.server(futures.ThreadPoolExecutor(max_workers=30))

    vizier_service_pb2_grpc.add_VizierServiceServicer_to_server(
        self._servicer, self._server)
    self._server.add_secure_port(self.endpoint, grpc.local_server_credentials())
    self._server.start()
    self.stub = vizier_client.create_server_stub(self.endpoint)

  def wait_for_early_stop_recycle_period(self) -> None:
    time.sleep(self._early_stop_recycle_period.total_seconds())



"""Experimenter base class for problem statement and evaluation.

Experimenters represent black-box optimization problems.
Each experimenter defines a ProblemStatement, representing the search space and
the metrics it returns in Evaluate (via CompletedTrials).

Pseudo-code for using Experimenters with Vizier Designers:

exp = ExperimenterSubClass(...)  # Possibly configure the experimenter.
problem_statement = exp.problem_statement()
designer = Designer(problem_statement)  # Configure the search algorithm
for i in range(10):
  suggestions = designer.suggest(count=2)
  completed_trials = exp.evaluate(suggestions)
  designer.update(completed_trials)
"""

import abc
from typing import List, Sequence

from vizier import pyvizier


class Experimenter(metaclass=abc.ABCMeta):
  """Abstract base class for Experimenters."""

  # TODO: Use CompletedTrials as output type.
  @abc.abstractmethod
  def evaluate(self,
               suggestions: Sequence[pyvizier.Trial]) -> List[pyvizier.Trial]:
    """Evaluates and completes the Trials.

    Args:
      suggestions: Sequence of Trials to be evaluated.

    Returns:
      List of completed Trials.
    """
    pass

  @abc.abstractmethod
  def problem_statement(self) -> pyvizier.ProblemStatement:
    """The search configuration generated by this experimenter."""
    pass



"""Tests for vizier.pyvizier.oss.metadata_util."""

from vizier._src.pyvizier.oss import metadata_util
from vizier.service import study_pb2

from absl.testing import absltest


class MetadataUtilTest(absltest.TestCase):

  def test_get(self):
    meta_trial = study_pb2.Trial(id='meta_trial')
    trial = study_pb2.Trial(id='trial')
    metadata_util.assign(trial, key='any', ns='', value=meta_trial)
    metadata_util.assign(trial, key='text', ns='x', value='x-value')
    metadata_util.assign(trial, key='text', ns='', value='value')
    metadata_util.assign(trial, key='text', ns='y', value='y-value')
    self.assertEqual(
        metadata_util.get_proto(trial, key='any', ns='', cls=study_pb2.Trial),
        meta_trial)
    self.assertEqual(metadata_util.get(trial, key='text', ns=''), 'value')
    self.assertEqual(metadata_util.get(trial, key='text', ns='x'), 'x-value')

    self.assertIsNone(
        metadata_util.get_proto(trial, key='any', ns='', cls=study_pb2.Study))
    self.assertIsNone(metadata_util.get(trial, key='TYPO', ns=''))
    self.assertIsNone(
        metadata_util.get_proto(trial, key='TYPO', ns='', cls=study_pb2.Trial))


if __name__ == '__main__':
  absltest.main()



"""Tests for vizier.service.vizier_server."""
# TODO: Change the test to create a vizier stub and call its
# methods, instead of directly calling VizierService methods.

import datetime
import time
from vizier.service import key_value_pb2
from vizier.service import resources
from vizier.service import study_pb2
from vizier.service import test_util
from vizier.service import vizier_server
from vizier.service import vizier_service_pb2

from google.longrunning import operations_pb2

from absl.testing import absltest
from absl.testing import parameterized

_KeyValuePlus = vizier_service_pb2.UpdateMetadataRequest.KeyValuePlus


class VizierServerTest(parameterized.TestCase):

  def setUp(self):
    self.early_stop_recycle_period = datetime.timedelta(seconds=0.1)
    self.vs = vizier_server.VizierService(
        early_stop_recycle_period=self.early_stop_recycle_period)
    self.owner_id = 'my_username'
    self.study_id = '0123123'
    self.client_id = 'client_0'
    super().setUp()

  @parameterized.named_parameters(
      ('Maximize', study_pb2.StudySpec.MetricSpec.MAXIMIZE),
      ('Minimize', study_pb2.StudySpec.MetricSpec.MINIMIZE))
  def test_single_objective_list_optimal(self, max_or_min_proto):
    metric_id = 'accuracy'
    max_or_min_study = test_util.generate_study(
        self.owner_id,
        self.study_id,
        study_spec=study_pb2.StudySpec(metrics=[
            study_pb2.StudySpec.MetricSpec(
                metric_id=metric_id, goal=max_or_min_proto)
        ]))

    self.vs.datastore.create_study(max_or_min_study)

    lowest_trial_final_measurement = study_pb2.Measurement(
        metrics=[study_pb2.Measurement.Metric(metric_id=metric_id, value=-1.0)])

    middle_trial_final_measurement = study_pb2.Measurement(
        metrics=[study_pb2.Measurement.Metric(metric_id=metric_id, value=0.0)])

    highest_trial_final_measurement = study_pb2.Measurement(
        metrics=[study_pb2.Measurement.Metric(metric_id=metric_id, value=1.0)])

    lowest_trial = test_util.generate_trials(
        trial_id_list=[1],
        owner_id=self.owner_id,
        study_id=self.study_id,
        final_measurement=lowest_trial_final_measurement,
        state=study_pb2.Trial.State.SUCCEEDED)[0]

    middle_trial = test_util.generate_trials(
        trial_id_list=[2],
        owner_id=self.owner_id,
        study_id=self.study_id,
        final_measurement=middle_trial_final_measurement,
        state=study_pb2.Trial.State.SUCCEEDED)[0]

    highest_trial = test_util.generate_trials(
        trial_id_list=[3],
        owner_id=self.owner_id,
        study_id=self.study_id,
        final_measurement=highest_trial_final_measurement,
        state=study_pb2.Trial.State.SUCCEEDED)[0]

    another_highest_trial = test_util.generate_trials(
        trial_id_list=[4],
        owner_id=self.owner_id,
        study_id=self.study_id,
        final_measurement=highest_trial_final_measurement,
        state=study_pb2.Trial.State.SUCCEEDED)[0]

    trials_to_be_ignored = test_util.generate_trials([5, 6, 7], self.owner_id,
                                                     self.study_id)

    all_trials = [
        lowest_trial, middle_trial, highest_trial, another_highest_trial
    ] + trials_to_be_ignored

    for trial in all_trials:
      self.vs.datastore.create_trial(trial)

    optimal_trial_list = self.vs.ListOptimalTrials(
        request=vizier_service_pb2.ListOptimalTrialsRequest(
            parent=max_or_min_study.name)).optimal_trials

    if max_or_min_study == study_pb2.StudySpec.MetricSpec.MAXIMIZE:
      self.assertLen(optimal_trial_list, 2)
    elif max_or_min_study == study_pb2.StudySpec.MetricSpec.MINIMIZE:
      self.assertEqual(optimal_trial_list[0], lowest_trial)

  def test_multiobjective_list_optimal(self):
    metric_id_1 = 'x1'
    metric_id_2 = 'x2'
    study = test_util.generate_study(
        self.owner_id,
        self.study_id,
        study_spec=study_pb2.StudySpec(metrics=[
            study_pb2.StudySpec.MetricSpec(
                metric_id=metric_id_1,
                goal=study_pb2.StudySpec.MetricSpec.MAXIMIZE),
            study_pb2.StudySpec.MetricSpec(
                metric_id=metric_id_2,
                goal=study_pb2.StudySpec.MetricSpec.MINIMIZE),
        ]))
    self.vs.datastore.create_study(study)

    low_x1 = study_pb2.Measurement.Metric(metric_id=metric_id_1, value=-1.0)
    low_x2 = study_pb2.Measurement.Metric(metric_id=metric_id_2, value=-1.0)
    high_x1 = study_pb2.Measurement.Metric(metric_id=metric_id_1, value=1.0)
    high_x2 = study_pb2.Measurement.Metric(metric_id=metric_id_2, value=1.0)

    low_low = study_pb2.Measurement(metrics=[low_x1, low_x2])
    low_high = study_pb2.Measurement(metrics=[low_x1, high_x2])
    high_low = study_pb2.Measurement(metrics=[high_x1, low_x2])
    high_high = study_pb2.Measurement(metrics=[high_x1, high_x2])

    incomplete_measurement = study_pb2.Measurement(metrics=[low_x1])

    ll_trial = test_util.generate_trials(
        trial_id_list=[1],
        owner_id=self.owner_id,
        study_id=self.study_id,
        final_measurement=low_low,
        state=study_pb2.Trial.State.SUCCEEDED)[0]
    lh_trial = test_util.generate_trials(
        trial_id_list=[2],
        owner_id=self.owner_id,
        study_id=self.study_id,
        final_measurement=low_high,
        state=study_pb2.Trial.State.SUCCEEDED)[0]
    hl_trial = test_util.generate_trials(
        trial_id_list=[3],
        owner_id=self.owner_id,
        study_id=self.study_id,
        final_measurement=high_low,
        state=study_pb2.Trial.State.SUCCEEDED)[0]
    hh_trial = test_util.generate_trials(
        trial_id_list=[4],
        owner_id=self.owner_id,
        study_id=self.study_id,
        final_measurement=high_high,
        state=study_pb2.Trial.State.SUCCEEDED)[0]

    missing_measurement_trials = test_util.generate_trials(
        trial_id_list=[5, 6, 7],
        owner_id=self.owner_id,
        study_id=self.study_id,
        final_measurement=incomplete_measurement,
        state=study_pb2.Trial.State.SUCCEEDED)

    all_trials = [ll_trial, lh_trial, hl_trial, hh_trial
                 ] + missing_measurement_trials
    for trial in all_trials:
      self.vs.datastore.create_trial(trial)

    optimal_trial_list = self.vs.ListOptimalTrials(
        request=vizier_service_pb2.ListOptimalTrialsRequest(
            parent=study.name)).optimal_trials

    self.assertLen(optimal_trial_list, 1)
    self.assertEqual(optimal_trial_list[0], hl_trial)

  def test_suggest_trials(self):
    suggestion_count = 10

    example_study_spec = test_util.generate_all_four_parameter_specs(
        algorithm=study_pb2.StudySpec.Algorithm.RANDOM_SEARCH)
    example_study = test_util.generate_study(
        self.owner_id, self.study_id, study_spec=example_study_spec)
    self.vs.datastore.create_study(example_study)

    request = vizier_service_pb2.SuggestTrialsRequest(
        parent=resources.StudyResource(self.owner_id, self.study_id).name,
        suggestion_count=suggestion_count,
        client_id=self.client_id)
    operation = self.vs.SuggestTrials(request)

    # Check if operation was stored in database.
    get_operation_request = operations_pb2.GetOperationRequest(
        name=resources.SuggestionOperationResource(self.owner_id,
                                                   self.client_id, 1).name)
    get_operation_output = self.vs.GetOperation(get_operation_request)
    self.assertEqual(operation, get_operation_output)

    # Check operation contents.
    suggest_trials_response = vizier_service_pb2.SuggestTrialsResponse.FromString(
        operation.response.value)
    self.assertLen(suggest_trials_response.trials, suggestion_count)
    for trial in suggest_trials_response.trials:
      self.assertLen(trial.parameters, 4)

  @parameterized.named_parameters(('IncludeFinalMeasurement', True),
                                  ('NoFinalMeasurement', False))
  def test_complete_trial(self, include_final_measurement: bool):
    metric_id = 'x'
    study = test_util.generate_study(
        self.owner_id,
        self.study_id,
        study_spec=study_pb2.StudySpec(metrics=[
            study_pb2.StudySpec.MetricSpec(
                metric_id=metric_id,
                goal=study_pb2.StudySpec.MetricSpec.MAXIMIZE)
        ]))
    self.vs.datastore.create_study(study)

    trial = test_util.generate_trials(
        trial_id_list=[1],
        owner_id=self.owner_id,
        study_id=self.study_id,
        state=study_pb2.Trial.State.ACTIVE)[0]
    self.vs.datastore.create_trial(trial)
    complete_trial_request = vizier_service_pb2.CompleteTrialRequest(
        name=trial.name)

    if include_final_measurement:
      final_measurement = study_pb2.Measurement(metrics=[
          study_pb2.Measurement.Metric(metric_id=metric_id, value=-1.0)
      ])
      complete_trial_request.final_measurement.CopyFrom(final_measurement)
      self.vs.CompleteTrial(complete_trial_request)
    else:
      with self.assertRaises(ValueError):
        # trial and request both do not contain measurements.
        self.vs.CompleteTrial(complete_trial_request)

  def test_early_stopping(self):
    example_study_spec = test_util.generate_all_four_parameter_specs(
        algorithm=study_pb2.StudySpec.Algorithm.RANDOM_SEARCH)
    example_study = test_util.generate_study(
        self.owner_id, self.study_id, study_spec=example_study_spec)
    self.vs.datastore.create_study(example_study)

    active_trials = test_util.generate_trials(
        trial_id_list=[1, 2, 3, 4],
        owner_id=self.owner_id,
        study_id=self.study_id,
        state=study_pb2.Trial.State.ACTIVE)
    for t in active_trials:
      self.vs.datastore.create_trial(t)
      request = vizier_service_pb2.CheckTrialEarlyStoppingStateRequest(
          trial_name=t.name)
      response = self.vs.CheckTrialEarlyStoppingState(request)
      # Since RandomPolicy picks a random ACTIVE trial to stop and current trial
      # t is the only ACTIVE trial, it should always stop.
      self.assertTrue(
          response.should_stop,
          msg=f'trial={t}, request={request}, response={response}')
      stop_trial_request = vizier_service_pb2.StopTrialRequest(name=t.name)
      new_t = self.vs.StopTrial(stop_trial_request)
      self.assertEqual(new_t.state, study_pb2.Trial.State.STOPPING)

    for t in active_trials:
      trial_resource = resources.TrialResource.from_name(t.name)
      operation_name = resources.EarlyStoppingOperationResource(
          trial_resource.owner_id, trial_resource.study_id,
          trial_resource.trial_id).name
      op = self.vs.datastore.get_early_stopping_operation(operation_name)
      self.assertTrue(op.should_stop)

    # After a while, the opeartion will be recycled, and `should_stop` is
    # defaulted to False. Since the trial is no longer active, RandomPolicy will
    # not consider it.
    time.sleep(self.early_stop_recycle_period.total_seconds())
    request = vizier_service_pb2.CheckTrialEarlyStoppingStateRequest(
        trial_name=active_trials[0].name)
    response = self.vs.CheckTrialEarlyStoppingState(request)
    self.assertFalse(response.should_stop)

  def test_update_metadata(self):
    # Construct a study.
    example_study_spec = test_util.generate_all_four_parameter_specs(
        algorithm=study_pb2.StudySpec.Algorithm.RANDOM_SEARCH)
    example_study = test_util.generate_study(
        self.owner_id, self.study_id, study_spec=example_study_spec)
    self.vs.datastore.create_study(example_study)
    active_trials = test_util.generate_trials(
        trial_id_list=[1, 2],
        owner_id=self.owner_id,
        study_id=self.study_id,
        state=study_pb2.Trial.State.ACTIVE)
    for t in active_trials:
      self.vs.datastore.create_trial(t)

    # Construct the request.
    study_metadata = _KeyValuePlus(
        k_v=key_value_pb2.KeyValue(key='a', ns='b', value='C'))
    trial_metadata = _KeyValuePlus(
        trial_id='1', k_v=key_value_pb2.KeyValue(key='d', ns='e', value='F'))
    request = vizier_service_pb2.UpdateMetadataRequest(
        name=resources.StudyResource(self.owner_id, self.study_id).name,
        metadata=[study_metadata, trial_metadata])
    # Send it to the server.
    response = self.vs.UpdateMetadata(request)
    # Check that there was no error.
    self.assertEmpty(response.error_details)


if __name__ == '__main__':
  absltest.main()



"""Wrapper classes for Trial protos and other messages in them.

Example usage:
  trial = Trial.from_proto(trial_proto)
  print('This trial's auc is: ', trial.final_measurement.metrics['auc'].value)
  print('This trial had parameter "n_hidden_layers": ',
        trial.parameters['n_hidden_layers'].value)
"""

import collections
from collections import abc as cabc
import copy
import datetime
import enum
from typing import Any, Dict, List, MutableMapping, Optional, Union, FrozenSet

from absl import logging
import attr
import numpy as np

from vizier._src.pyvizier.shared import common

ParameterValueTypes = Union[str, int, float, bool]
OrderedDict = collections.OrderedDict
Metadata = common.Metadata


class ExternalType(enum.Enum):
  """Valid Values for ParameterConfig.external_type."""
  INTERNAL = 'INTERNAL'
  BOOLEAN = 'BOOLEAN'
  INTEGER = 'INTEGER'
  FLOAT = 'FLOAT'


# Values should NEVER be removed from the enums below, only added.
class TrialStatus(enum.Enum):
  """Values for Trial.Status."""
  UNKNOWN = 'UNKNOWN'
  REQUESTED = 'REQUESTED'
  ACTIVE = 'ACTIVE'
  COMPLETED = 'COMPLETED'
  STOPPING = 'STOPPING'


@attr.s(frozen=True, init=True, slots=True, kw_only=False)
class Metric:
  """Enhanced immutable wrapper for vizier_pb2.Metric proto.

  It has an additional field "std" for internal usage. This field gets lost
  when the object is converted to proto.
  """

  def _std_not_negative(self, _, stddev):
    if stddev < 0:
      raise ValueError(
          'Standard deviation must be a non-negative finite number.')

  value: float = attr.ib(
      converter=float,
      init=True,
      validator=[attr.validators.instance_of(float)],
      kw_only=False)
  std: float = attr.ib(
      converter=float,
      validator=[attr.validators.instance_of(float), _std_not_negative],
      init=True,
      default=0.0,
      kw_only=True)


# Use when you want to preserve the shapes or reduce if-else statements.
# e.g. `metrics.get('metric_name', NaNMetric).value` to get NaN or the actual
# value.
NaNMetric = Metric(value=np.nan)


@attr.s(auto_attribs=True, frozen=True, init=True, slots=True, repr=False)
class ParameterValue:
  """Immutable wrapper for vizier_pb2.Parameter.value, which is a oneof field.

  Has accessors (properties) that cast the value into the type according
  to StudyConfiguration class behavior. In particular, 'true' and 'false' are
  treated as special strings that are cast to a numeric value of 1 and 0,
  respectively, and boolean value of True and False, repectively.
  """

  value: ParameterValueTypes = attr.ib(
      init=True,
      validator=[
          attr.validators.instance_of((str, int, float, bool)),
      ])

  def cast(
      self,
      external_type: ExternalType,
  ) -> ParameterValueTypes:
    """Returns ParameterValue cast to external_type.

    Args:
      external_type:

    Returns:
      self.value if external_type is INTERNAL.
      self.as_bool if external_type is BOOLEAN.
      self.as_int if external_type is INTEGER.
      self.as_float if external_type is FLOAT.

    Raises:
      ValueError: If external_type is not valid.
    """
    if external_type == ExternalType.INTERNAL:
      return self.value
    elif external_type == ExternalType.BOOLEAN:
      return self.as_bool
    elif external_type == ExternalType.INTEGER:
      return self.as_int
    elif external_type == ExternalType.FLOAT:
      return self.as_float
    else:
      raise ValueError(
          'Unknown external type enum value: {}.'.format(external_type))

  @property
  def as_float(self) -> Optional[float]:
    """Returns the value cast to float."""
    if self.value == 'true':
      return 1.0
    elif self.value == 'false':
      return 0.0
    try:
      # Note str -> float conversion exists for benchmark use.
      return float(self.value)
    except ValueError:
      return None

  @property
  def as_int(self) -> Optional[int]:
    """Returns the value cast to int."""
    if self.value == 'true':
      return 1
    elif self.value == 'false':
      return 0
    try:
      # Note str -> int conversion exists for benchmark use.
      return int(self.value)
    except ValueError:
      return None

  @property
  def as_str(self) -> Optional[str]:
    """Returns str-typed value or lowercase 'true'/'false' if value is bool."""
    if isinstance(self.value, bool):
      return str(self.value).lower()
    elif isinstance(self.value, str):
      return self.value
    # Note str conversion exists for benchmark use. Use __str__ instead.
    return str(self.value)

  @property
  def as_bool(self) -> Optional[bool]:
    """Returns the value as bool following StudyConfiguration's behavior.

    Returns: True if value is 'true' or 1. False if value is
      'false' or 0. For all other cases, returns None.
      For string type, this behavior is consistent with how
      StudyConfiguration.AddBooleanParameter's. For other types, this
      guarantees that self.value == self.as_bool
    """
    if isinstance(self.value, str):
      if self.value.lower() == 'true':
        return True
      elif self.value.lower() == 'false':
        return False
    else:
      if self.value == 1.0:
        return True
      elif self.value == 0.0:
        return False
    return None

  def __str__(self) -> str:
    return str(self.value)

  def __repr__(self) -> str:
    return str(self.value)


class _MetricDict(collections.UserDict[str, Metric]):
  """Dictionary of string to metrics."""

  def get_value(self, key: str, default: float) -> float:
    if key in self.data:
      return self.data[key].value
    else:
      return default

  def __setitem__(self, key: str, value: Union[float, Metric]):
    if isinstance(value, Metric):
      self.data.__setitem__(key, value)
    else:
      self.data.__setitem__(key, Metric(value=value))


@attr.s(auto_attribs=True, frozen=False, init=True, slots=True)
class Measurement:
  """Collection of metrics with a timestamp."""

  def _value_is_finite(self, _, value):
    if not (np.isfinite(value) and value >= 0):
      raise ValueError('Must be finite and non-negative.')

  # Should be used as a regular Dict.
  metrics: _MetricDict = attr.ib(
      init=True,
      converter=lambda d: _MetricDict(**d),
      default=_MetricDict(),
      validator=attr.validators.instance_of(_MetricDict),
      on_setattr=[attr.setters.convert, attr.setters.validate])

  elapsed_secs: float = attr.ib(
      converter=float,
      init=True,
      default=0,
      validator=[attr.validators.instance_of(float), _value_is_finite],
      on_setattr=[attr.setters.convert, attr.setters.validate],
      kw_only=True)

  # TODO: Change type annotation to int.
  steps: float = attr.ib(
      converter=int,
      init=True,
      default=0,
      validator=[attr.validators.instance_of(int), _value_is_finite],
      on_setattr=[attr.setters.convert, attr.setters.validate],
      kw_only=True)


def _to_local_time(
    dt: Optional[datetime.datetime]) -> Optional[datetime.datetime]:
  """Converter for initializing timestamps in Trial class."""
  return dt.astimezone() if dt else None


@attr.define(init=False, frozen=True, eq=True)
class ParameterDict(cabc.MutableMapping):
  """Parameter dictionary.

  Maps the parameter names to their values. Works like a regular
  dict[str, ParameterValue] for the most part, except one can directly assign
  values of type `ParameterValueType`. So,
    ParameterDict(a=3) and
    ParameterDict(a=ParameterValue(3)) are equivalent.


  To access the raw value directly, use get_value() or as_dict():
    d.get_value('a') == d.get('a').value
    d.as_dict()['a'] == d.get_value('a')
  """

  _items: MutableMapping[str, ParameterValue] = attr.field(
      init=False, factory=dict)

  def as_dict(self) -> dict[str, ParameterValueTypes]:
    """Returns the dict of parameter names to raw values."""
    return {k: self.get_value(k) for k in self._items}

  def __init__(self, iterable: Any = tuple(), **kwargs):
    self.__attrs_init__()
    self.update(iterable, **kwargs)

  def __setitem__(self, key: str, value: Union[ParameterValue,
                                               ParameterValueTypes]):
    if isinstance(value, ParameterValue):
      self._items[key] = value
    else:
      self._items[key] = ParameterValue(value)

  def __delitem__(self, key: str):
    del self._items[key]

  def __getitem__(self, key: str) -> ParameterValue:
    return self._items[key]

  def __len__(self) -> int:
    return len(self._items)

  def __iter__(self):
    return iter(self._items)

  def get_value(
      self,
      key: str,
      default: Optional[ParameterValueTypes] = None
  ) -> Optional[ParameterValueTypes]:
    """Returns the raw value of the given parameter name."""
    pv = self.get(key, default)
    if isinstance(pv, ParameterValue):
      return pv.value
    else:
      return pv


@attr.define(auto_attribs=True, frozen=False, init=True, slots=True)
class TrialSuggestion:
  """Freshly suggested trial.

  Suggestion can be converted to Trial object which has more functionalities.
  """

  parameters: ParameterDict = attr.field(
      init=True,
      factory=ParameterDict,
      converter=ParameterDict,
      validator=attr.validators.instance_of(ParameterDict))  # pytype: disable=wrong-arg-types

  metadata: Metadata = attr.field(
      init=True,
      kw_only=True,
      factory=Metadata,
      validator=attr.validators.instance_of(Metadata))

  def to_trial(self, uid: int = 0) -> 'Trial':
    """Assign an id and make it a Trial object.

    Usually SuggetedTrial objects are shorted-lived and not exposed to end
    users. This method is for non-service usage of trial suggestions in
    benchmarks, tests, colabs, etc.

    Args:
      uid: Trial id.

    Returns:
      Trial object.
    """
    return Trial(id=uid, parameters=self.parameters, metadata=self.metadata)


@attr.define(auto_attribs=True, frozen=False, init=True, slots=True)
class Trial(TrialSuggestion):
  """Wrapper for learning_vizier.service.Trial proto."""
  id: int = attr.ib(
      init=True,
      kw_only=True,
      default=0,
      validator=attr.validators.instance_of(int),
  )

  _is_requested: bool = attr.ib(
      init=True,
      kw_only=True,
      default=False,
      validator=attr.validators.instance_of(bool))

  assigned_worker: Optional[str] = attr.ib(
      init=True,
      kw_only=True,
      default=None,
      validator=attr.validators.optional(attr.validators.instance_of(str)),
  )

  stopping_reason: Optional[str] = attr.ib(
      init=True,
      kw_only=True,
      default=None,
      validator=attr.validators.optional(attr.validators.instance_of(str)),
  )

  _infeasibility_reason: Optional[str] = attr.ib(
      init=True,
      kw_only=True,
      default=None,
      validator=attr.validators.optional(attr.validators.instance_of(str)),
  )

  description: Optional[str] = attr.ib(
      init=True,
      kw_only=True,
      default=None,
      validator=attr.validators.optional(attr.validators.instance_of(str)),
  )

  related_links: Dict[str, str] = attr.ib(
      init=True,
      kw_only=True,
      factory=dict,
      validator=attr.validators.deep_mapping(
          key_validator=attr.validators.instance_of(str),
          value_validator=attr.validators.instance_of(str),
          mapping_validator=attr.validators.instance_of(dict)),
  )  # pytype: disable=wrong-arg-types

  final_measurement: Optional[Measurement] = attr.ib(
      init=True,
      kw_only=True,
      default=None,
      validator=attr.validators.optional(
          attr.validators.instance_of(Measurement)),
  )

  measurements: List[Measurement] = attr.ib(
      init=True,
      kw_only=True,
      factory=list,
      validator=attr.validators.deep_iterable(
          member_validator=attr.validators.instance_of(Measurement),
          iterable_validator=attr.validators.instance_of(list)),
  )

  creation_time: Optional[datetime.datetime] = attr.ib(
      init=True,
      default=datetime.datetime.now(),
      converter=_to_local_time,
      kw_only=True,
      repr=lambda v: v.strftime('%x %X') if v is not None else 'None',
      validator=attr.validators.optional(
          attr.validators.instance_of(datetime.datetime)),
  )

  completion_time: Optional[datetime.datetime] = attr.ib(
      init=True,
      kw_only=True,
      default=None,
      repr=lambda v: v.strftime('%x %X') if v is not None else 'None',
      converter=_to_local_time,
      validator=attr.validators.optional(
          attr.validators.instance_of(datetime.datetime)),
  )

  @property
  def duration(self) -> Optional[datetime.timedelta]:
    """Returns the duration of this Trial if it is completed, or None."""
    if self.completion_time:
      return self.completion_time - self.creation_time
    else:
      return None

  @property
  def status(self) -> TrialStatus:
    """Status.

    COMPLETED: Trial has final measurement or is declared infeasible.
    ACTIVE: Trial is being evaluated.
    STOPPING: Trial is being evaluated, but was decided to be not worth further
      evaluating.
    REQUESTED: Trial is queued for future suggestions.
    """
    if self.final_measurement is not None or self.infeasible:
      return TrialStatus.COMPLETED
    elif self.stopping_reason is not None:
      return TrialStatus.STOPPING
    elif self._is_requested:
      return TrialStatus.REQUESTED
    else:
      return TrialStatus.ACTIVE

  @property
  def is_completed(self) -> bool:
    """Returns True if this Trial is completed."""
    if self.status == TrialStatus.COMPLETED:
      if self.completion_time is None:
        logging.warning('Invalid Trial state: status is COMPLETED, but a '
                        ' completion_time was not set')
      return True
    elif self.completion_time is not None:
      if self.status is None:
        logging.warning('Invalid Trial state: status is not set to COMPLETED, '
                        'but a completion_time is set')
      return True
    return False

  @property
  def infeasible(self) -> bool:
    """Returns True if this Trial is infeasible."""
    return self._infeasibility_reason is not None

  @property
  def infeasibility_reason(self) -> Optional[str]:
    """Returns this Trial's infeasibility reason, if set."""
    return self._infeasibility_reason

  def complete(self,
               measurement: Measurement,
               *,
               inplace: bool = True) -> 'Trial':
    """Completes the trial and returns it.

    Args:
      measurement: Measurement to complete the trial with.
      inplace: If True, Trial is modified in place. If False,  which is the
        default, then the operation is performed and it returns a copy of the
        object

    Returns:
      Completed Trial.
    """
    if inplace:
      # Use setattr. If we assign to self.final_measurement, then hyperref
      # mechanisms think this line is where `final_measurement` property
      # is defined, instead of where we declare attr.ib.
      self.__setattr__('final_measurement', copy.deepcopy(measurement))
      self.completion_time = _to_local_time(datetime.datetime.now())
      return self
    else:
      clone = copy.deepcopy(self)
      return clone.complete(measurement, inplace=True)


# Define aliases.
CompletedTrial = Trial
PendingTrial = Trial
CompletedTrialWithMeasurements = Trial
PendingTrialWithMeasurements = Trial


@attr.define(kw_only=True)
class TrialFilter:
  """Trial filter.

  All filters are by default 'AND' conditions.

  Attributes:
    ids: If set, requires the trial's id to be in the set.
    min_id: If set, requires the trial's id to be at least this number.
    max_id: If set, requires the trial's id to be at most this number.
    status: If set, requires the trial's status to be in the set.
  """
  ids: Optional[FrozenSet[int]] = attr.field(
      default=None,
      converter=lambda x: frozenset(x) if x is not None else None,
      validator=attr.validators.optional(
          attr.validators.deep_iterable(
              attr.validators.instance_of(int),
              attr.validators.instance_of(frozenset))))
  min_id: Optional[int] = attr.field(default=None)
  max_id: Optional[int] = attr.field(default=None)
  status: Optional[FrozenSet[TrialStatus]] = attr.field(
      default=None,
      converter=lambda x: frozenset(x) if x is not None else None,
      validator=attr.validators.optional(
          attr.validators.deep_iterable(
              attr.validators.instance_of(TrialStatus),
              attr.validators.instance_of(frozenset))))

  # TODO: Add "search_space" argument

  def __call__(self, trial: Trial) -> bool:
    if self.ids is not None:
      if trial.id not in self.ids:
        return False
    if self.min_id is not None:
      if trial.id < self.min_id:
        return False
    if self.max_id is not None:
      if trial.id > self.max_id:
        return False
    if self.status is not None:
      if trial.status not in self.status:
        return False
    return True



"""Numpy experimenter for wrapping deterministic functions on ndarrays."""

import copy
import logging
import math
from typing import Callable, List, Sequence

import numpy as np
from vizier import pyvizier
from vizier._src.benchmarks.experimenters import experimenter
from vizier.pyvizier import converters


def _get_name(f):
  """Gets the name of underlying objects."""
  if hasattr(f, '__name__'):
    return f.__name__
  # Next clause handles functools.partial objects.
  if hasattr(f, 'func') and hasattr(f.func, '__name__'):
    return f.func.__name__
  return repr(f)


class NumpyExperimenter(experimenter.Experimenter):
  """NumpyExperimenters take a deterministic function on ndarrays."""

  def __init__(self, impl: Callable[[np.ndarray], float],
               problem_statement: pyvizier.ProblemStatement):
    """NumpyExperimenter with analytic function impl for one metric.

    NumpyExperimenter only supports single objectives, and flat numeric search
    spaces.

    Args:
      impl: Function that scalarizes np.ndarray of shape (dimension,).
      problem_statement: Problem statement.

    Raises:
      ValueError: Non-positive dimension or invalid problem statement.
    """
    dimension = len(problem_statement.search_space.parameters)
    logging.info('Initializing NumpyExperimenter with impl=%s, dimension=%s',
                 _get_name(impl), dimension)
    if dimension <= 0:
      raise ValueError(f'Invalid dimension: {dimension}')
    self.impl = impl

    if not problem_statement.metric_information.is_single_objective:
      raise ValueError(
          f'Statement should be single objective {problem_statement}')
    if problem_statement.search_space.is_conditional:
      raise ValueError(f'Statement should be flat {problem_statement}')
    for parameter in problem_statement.search_space.parameters:
      if not parameter.type.is_numeric():
        raise ValueError(f'Non-numeric parameters {parameter}')

    self._metric_name = problem_statement.metric_information.of_type(
        pyvizier.MetricType.OBJECTIVE).item().name
    self._problem_statement = copy.deepcopy(problem_statement)
    self._converter = converters.TrialToArrayConverter.from_study_config(
        study_config=self._problem_statement,
        scale=False,
        flip_sign_for_minimization_metrics=False)

  def problem_statement(self) -> pyvizier.ProblemStatement:
    return copy.deepcopy(self._problem_statement)

  def evaluate(self,
               suggestions: Sequence[pyvizier.Trial]) -> List[pyvizier.Trial]:
    completed_trials = list(copy.deepcopy(suggestions))
    # Features has shape (num_trials, num_features).
    features = self._converter.to_features(completed_trials)
    for idx, completed_trial in enumerate(completed_trials):
      val = self.impl(features[idx])
      if math.isfinite(val):
        completed_trial.complete(
            pyvizier.Measurement(metrics={self._metric_name: val}))
      else:
        # TODO: Add infeasibility to completed_trial.
        completed_trial.complete(pyvizier.Measurement())
      if not completed_trial.is_completed:
        raise RuntimeError(f'Trial {completed_trial} not completed')
    return completed_trials

  def __repr__(self) -> str:
    return f'NumpyExperimenter {{name: {_get_name(self.impl)}'



"""Import target for multimetric."""

from vizier._src.pyvizier.multimetric.hypervolume import ParetoFrontier
from vizier._src.pyvizier.multimetric.pareto_optimal import FastParetoOptimalAlgorithm
from vizier._src.pyvizier.multimetric.pareto_optimal import NaiveParetoOptimalAlgorithm
from vizier._src.pyvizier.multimetric.safety import SafetyChecker



"""Library functions for testing databases."""
import copy
import random
import string
from typing import List

from vizier.service import datastore
from vizier.service import key_value_pb2
from vizier.service import resources
from vizier.service import study_pb2
from vizier.service import vizier_oss_pb2
from vizier.service import vizier_service_pb2

from google.longrunning import operations_pb2
from absl.testing import parameterized

_KeyValuePlus = vizier_service_pb2.UpdateMetadataRequest.KeyValuePlus


def make_random_string() -> str:
  return ''.join(random.choice(string.ascii_lowercase) for _ in range(10))


class DataStoreTestCase(parameterized.TestCase):
  """Base class for testing datastores."""

  def assertStudyAPI(self, ds: datastore.DataStore, study: study_pb2.Study):
    """Tests if the datastore handles studies correctly."""
    ds.create_study(study)
    with self.assertRaises(datastore.AlreadyExistsError):
      ds.create_study(study)  # Can't make another study w/ same name.

    copied_study = ds.load_study(study.name)
    with self.assertRaises(datastore.NotFoundError):
      ds.load_study(study.name + 'does_not_exist')  # Non-existent study.
    self.assertEqual(copied_study, study)
    self.assertIsNot(copied_study, study)  # Check pass-by-value.

    owner_name = resources.StudyResource.from_name(
        study.name).owner_resource.name
    list_of_one_study = ds.list_studies(owner_name)
    with self.assertRaises(datastore.NotFoundError):
      ds.list_studies(owner_name + 'does_not_exist')  # Non-existent study.
    self.assertLen(list_of_one_study, 1)
    self.assertEqual(list_of_one_study[0], study)
    self.assertIsNot(list_of_one_study[0], study)  # Check pass-by-value.

    study.inactive_reason = make_random_string()
    copied_original_study = ds.load_study(study.name)
    self.assertNotEqual(study, copied_original_study)  # Check pass-by-value.

    ds.delete_study(study.name)
    with self.assertRaises(datastore.NotFoundError):
      ds.delete_study(study.name)  # Deleting non-existent study.
    with self.assertRaises(datastore.NotFoundError):
      ds.load_study(study.name)  # Non-existent study.

    # Owner should be kept track of.
    expected_studies = ds.list_studies(owner_name)
    self.assertEmpty(expected_studies)

  def assertTrialAPI(self, ds: datastore.DataStore, study: study_pb2.Study,
                     trials: List[study_pb2.Trial]):
    """Tests if the datastore handles trials correctly."""
    num_trials = len(trials)

    ds.create_study(study)
    empty_list = ds.list_trials(study.name)
    self.assertEmpty(empty_list)

    expected_max_trial_id = ds.max_trial_id(study.name)
    self.assertEqual(expected_max_trial_id, 0)

    for trial in trials:
      ds.create_trial(trial)
      with self.assertRaises(datastore.AlreadyExistsError):
        ds.create_trial(trial)  # Already exists.
      copied_trial = ds.get_trial(trial.name)
      with self.assertRaises(datastore.NotFoundError):
        ds.get_trial(trial.name + str(num_trials))  # Does not exist.
      self.assertEqual(trial, copied_trial)
      self.assertIsNot(trial, copied_trial)  # Check pass-by-value.

    self.assertLen(trials, ds.max_trial_id(study.name))
    with self.assertRaises(datastore.NotFoundError):
      ds.max_trial_id(study.name + 'does_not_exist')  # Does not exist.

    list_of_trials = ds.list_trials(study.name)
    with self.assertRaises(datastore.NotFoundError):
      ds.list_trials(study.name + 'does_not_exist')  # Does not exist.
    self.assertEqual(list_of_trials, trials)
    self.assertIsNot(list_of_trials, trials)  # Check pass-by-value.

    first_trial = trials[0]
    first_trial.infeasible_reason = make_random_string()
    ds.update_trial(first_trial)
    with self.assertRaises(datastore.NotFoundError):
      missing_trial = copy.deepcopy(first_trial)
      missing_trial.name += str(num_trials)
      ds.update_trial(missing_trial)  # Does not exist.
    new_first_trial = ds.get_trial(first_trial.name)
    self.assertEqual(first_trial, new_first_trial)
    self.assertIsNot(first_trial, new_first_trial)  # Check pass-by-value.

    ds.delete_trial(first_trial.name)
    with self.assertRaises(datastore.NotFoundError):
      ds.delete_trial(first_trial.name)  # Already deleted.
      ds.delete_trial(first_trial.name + str(num_trials))  # Does not exist.
    leftover_trials = ds.list_trials(study.name)
    self.assertEqual(leftover_trials, trials[1:])
    self.assertIsNot(leftover_trials, trials[1:])  # Check pass-by-value.

  def assertSuggestOpAPI(self, ds: datastore.DataStore, study: study_pb2.Study,
                         client_id: str,
                         suggestion_ops: List[operations_pb2.Operation]):
    """Tests if the datastore handles suggest ops correctly."""
    study_resource = resources.StudyResource.from_name(study.name)

    ds.create_study(study)
    for operation in suggestion_ops:
      ds.create_suggestion_operation(operation)
      with self.assertRaises(datastore.AlreadyExistsError):
        ds.create_suggestion_operation(operation)  # Already exists.

    owner_name = resources.OwnerResource(study_resource.owner_id).name
    self.assertLen(suggestion_ops,
                   ds.max_suggestion_operation_number(owner_name, client_id))

    with self.assertRaises(datastore.NotFoundError):
      ds.max_suggestion_operation_number(
          owner_name, client_id + 'does_not_exist')  # Client doesn't exist.

    list_of_operations = ds.list_suggestion_operations(owner_name, client_id)
    self.assertEqual(list_of_operations, suggestion_ops)
    with self.assertRaises(datastore.NotFoundError):
      ds.list_suggestion_operations(owner_name, client_id +
                                    'does_not_exist')  # Client doesn't exist.

    output_op = ds.get_suggestion_operation(
        resources.SuggestionOperationResource(
            study_resource.owner_id, client_id, operation_number=1).name)
    self.assertEqual(output_op, suggestion_ops[0])
    self.assertIsNot(output_op, suggestion_ops[0])  # Check pass-by-value.

    with self.assertRaises(datastore.NotFoundError):
      ds.get_suggestion_operation(
          resources.SuggestionOperationResource(
              study_resource.owner_id,
              client_id + 'does_not_exist',  # Client doesn't exist.
              operation_number=1).name)

    output_op.metadata.type_url = make_random_string()
    ds.update_suggestion_operation(output_op)
    new_output_op = ds.get_suggestion_operation(output_op.name)
    self.assertEqual(output_op, new_output_op)

    wrong_output_op = copy.deepcopy(output_op)
    wrong_output_op.name = resources.SuggestionOperationResource(
        study_resource.owner_id,
        client_id + 'does_not_exist',  # Client doesn't exist.
        operation_number=1).name
    with self.assertRaises(datastore.NotFoundError):
      ds.update_suggestion_operation(wrong_output_op)

  def assertEarlyStoppingAPI(
      self, ds: datastore.DataStore, study: study_pb2.Study,
      trials: List[study_pb2.Trial],
      early_stopping_ops: List[vizier_oss_pb2.EarlyStoppingOperation]):
    """Tests if the datastore handles early stopping ops correctly."""
    study_resource = resources.StudyResource.from_name(study.name)
    ds.create_study(study)

    for trial in trials:
      ds.create_trial(trial)

    for operation in early_stopping_ops:
      ds.create_early_stopping_operation(operation)
      with self.assertRaises(datastore.AlreadyExistsError):
        ds.create_early_stopping_operation(operation)  # Op already exists.

    output_op = ds.get_early_stopping_operation(
        resources.EarlyStoppingOperationResource(study_resource.owner_id,
                                                 study_resource.study_id,
                                                 1).name)
    self.assertEqual(output_op, early_stopping_ops[0])
    self.assertIsNot(output_op, early_stopping_ops[0])  # Check pass-by-value.

    wrong_op_name = resources.EarlyStoppingOperationResource(
        study_resource.owner_id,
        study_resource.study_id + 'does_not_exist',  # Study doesn't exist.
        1).name
    with self.assertRaises(datastore.NotFoundError):
      ds.get_early_stopping_operation(wrong_op_name)

    output_op.failure_message = make_random_string()
    ds.update_early_stopping_operation(output_op)
    new_output_op = ds.get_early_stopping_operation(output_op.name)
    self.assertEqual(output_op, new_output_op)

    wrong_output_op = copy.deepcopy(output_op)
    wrong_output_op.name = resources.EarlyStoppingOperationResource(
        study_resource.owner_id,
        study_resource.study_id + 'does_not_exist',  # Study doesn't exist.
        1).name
    with self.assertRaises(datastore.NotFoundError):
      ds.update_early_stopping_operation(wrong_output_op)

  def assertUpdateMetadataAPI(self, ds: datastore.DataStore,
                              study: study_pb2.Study,
                              trials: List[study_pb2.Trial]):
    """Tests if the datastore handles metadata updates properly."""
    # TODO: Make this test more general.
    ds.create_study(study)
    for trial in trials:
      ds.create_trial(trial)
    study_metadata = [key_value_pb2.KeyValue(key='a', ns='b', value='C')]
    trial_metadata = [
        _KeyValuePlus(
            trial_id='1',
            k_v=key_value_pb2.KeyValue(key='d', ns='e', value='F'))
    ]
    ds.update_metadata(study.name, study_metadata, trial_metadata)
    mutated_study_config = ds.load_study(study.name).study_spec
    self.assertEqual(list(mutated_study_config.metadata), study_metadata)
    mutated_trial = ds.get_trial(trials[0].name)
    self.assertEqual(mutated_trial.id, str(trial_metadata[0].trial_id))
    self.assertEqual(list(mutated_trial.metadata), [trial_metadata[0].k_v])



"""Init file."""
import os
import sys

THIS_DIR = os.path.dirname(os.path.realpath(__file__))
PROTO_ROOT = os.path.realpath(os.path.join(THIS_DIR, "service"))

sys.path.append(PROTO_ROOT)

__version__ = "0.0.2.alpha"



"""Random Pythia Policy which produces uniform sampling of Trial parameter values.

Since this is a RandomPolicy (i.e. stateless), we don't use the PolicySupporter
API when suggesting trials, but we do for the early stopping in order to
showcase how the policy supporter should be used.
"""
import random
from typing import List
from vizier import pythia
from vizier import pyvizier


def sample_from_parameter_config(
    parameter_config: pyvizier.ParameterConfig) -> pyvizier.ParameterValue:
  """Generate random values uniformly from a parameter_config."""
  if parameter_config.type == pyvizier.ParameterType.DOUBLE:
    min_value, max_value = parameter_config.bounds
    parameter_value = pyvizier.ParameterValue(
        value=random.uniform(min_value, max_value))
  elif parameter_config.type == pyvizier.ParameterType.INTEGER:
    min_value, max_value = parameter_config.bounds
    parameter_value = pyvizier.ParameterValue(
        value=random.randint(min_value, max_value))
  elif parameter_config.type == pyvizier.ParameterType.CATEGORICAL:
    parameter_value = pyvizier.ParameterValue(
        value=random.choice(parameter_config.feasible_values))
  elif parameter_config.type == pyvizier.ParameterType.DISCRETE:
    parameter_value = pyvizier.ParameterValue(
        value=random.choice(parameter_config.feasible_values))
  else:
    raise ValueError(
        f'ParameterConfig type is not one of the supported primitives for ParameterConfig: {parameter_config}'
    )
  return parameter_value


def make_random_parameters(
    study_config: pyvizier.StudyConfig) -> pyvizier.ParameterDict:
  """Makes random parameters from study_spec."""
  # TODO: Add conditional sampling case.
  parameter_dict = pyvizier.ParameterDict()
  for parameter_config in study_config.search_space.parameters:
    p_value = sample_from_parameter_config(parameter_config)
    parameter_dict[parameter_config.name] = p_value
  return parameter_dict


class RandomPolicy(pythia.Policy):
  """A policy that picks random hyper-parameter values."""

  def __init__(self, policy_supporter: pythia.PolicySupporter):
    self._policy_supporter = policy_supporter

  def suggest(self, request: pythia.SuggestRequest) -> pythia.SuggestDecisions:
    """Gets number of Trials to propose, and produces random Trials."""
    suggest_decision_list = []
    for _ in range(request.count):
      parameters = make_random_parameters(request.study_config)
      suggest_decision_list.append(
          pythia.SuggestDecision(parameters=parameters))
    return pythia.SuggestDecisions(suggest_decision_list)

  def early_stop(
      self, request: pythia.EarlyStopRequest) -> List[pythia.EarlyStopDecision]:
    """Selects a random ACTIVE/PENDING trial to stop from datastore."""
    early_stop_decisions = []

    all_active_trials = self._policy_supporter.GetTrials(
        study_guid=request.study_guid,
        status_matches=pyvizier.TrialStatus.ACTIVE)
    trial_to_stop_id = None
    if all_active_trials:
      trial_to_stop_id = random.choice(all_active_trials).id
      early_stop_decisions.append(
          pythia.EarlyStopDecision(
              id=trial_to_stop_id, reason='Random early stopping.'))

    for trial_id in list(request.trial_ids):
      if trial_id != trial_to_stop_id:
        early_stop_decisions.append(
            pythia.EarlyStopDecision(
                id=trial_id, reason='Trial should not stop.',
                should_stop=False))

    return early_stop_decisions
